#
#    This file is part of PyNomo -
#    a program to create nomographs with Python (http://pynomo.sourceforge.net/)
#
#    Copyright (C) 2007-2009  Leif Roschier  <lefakkomies@users.sourceforge.net>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

#from nomo_axis import *
from .nomo_axis import Nomo_Axis
import pyx
import numpy as np

class Circ_Scale:
    """
    class for primitive circular scales
    """

    def __init__(self, canvas, circ_appear={}):
        circ_appear_default_values = {
            'function': lambda u: u,
            'u_min': 1.0,
            'u_max': 10.0,
            'angle_min': 0.0,
            'angle_max': 180.0,
            'radius': 12,
            'angle_tick_direction': 'inner',  # or 'outer'
            'tick_levels': 4,
            'tick_text_levels': 2,
            'scale_type': 'linear',
            'text_distance_4': 1.0 / 4,
            'grid_length': 0.1,
            'grid_length_0': 0.5 / 4,
            'grid_length_1': 0.5 / 4,
            'grid_length_2': 0.5 / 4,
            'grid_length_3': 0.4 / 4,
            'grid_length_4': 0.3 / 4,
            'text_size': pyx.text.size.scriptsize,
            'text_size_0': pyx.text.size.tiny,
            'text_size_1': pyx.text.size.tiny,
            'text_size_2': pyx.text.size.tiny,
            'text_size_3': pyx.text.size.tiny,
            'text_size_4': pyx.text.size.tiny,
            'text_size_log_0': pyx.text.size.tiny,
            'text_size_log_1': pyx.text.size.tiny,
            'text_size_log_2': pyx.text.size.tiny,
            'text_size_manual': pyx.text.size.tiny,
            'text_distance_0': 1.0 / 4,
            'text_distance_1': 1.0 / 4,
            'text_distance_2': 1.0 / 4,
            'text_distance_3': 1.0 / 4,
            'text_distance_4': 1.0 / 4,
            'title_distance_center': 0.5,
            'title_opposite_tick': True,
            'title_draw_center': False,
            'text_format': "$%3.1f$",
            'full_angle': True,
            'extra_angle': 90.0,
            'text_horizontal_align_center': True,
            'manual_axis_data': {},
            'circ_scaling': None}
        self.circ_appear = circ_appear_default_values
        self.circ_appear.update(circ_appear)
        self.canvas = canvas
        self._make_f_and_g_func_()

    def draw(self):
        """
        draws the circular scale
        """
        Nomo_Axis(func_f=self.func_f, func_g=self.func_g,
                  start=self.circ_appear['u_min'], stop=self.circ_appear['u_max'],
                  turn=-1, title='circ',
                  tick_levels=self.circ_appear['tick_levels'],
                  tick_text_levels=self.circ_appear['tick_text_levels'],
                  canvas=self.canvas, type=self.circ_appear['scale_type'],
                  manual_axis_data=self.circ_appear['manual_axis_data'],
                  side=self.side, axis_appear=self.circ_appear)

    def _make_f_and_g_func_(self):
        """
        defines f and g functions. Assumes monotonic functions.
        """
        func = self.circ_appear['function']
        radius = self.circ_appear['radius']
        u_min = self.circ_appear['u_min']
        u_max = self.circ_appear['u_max']
        angle_min = self.circ_appear['angle_min']
        angle_max = self.circ_appear['angle_max']
        if self.circ_appear['circ_scaling'] == None:
            scaling = (angle_max - angle_min) / (func(u_max) - func(u_min)) * np.pi / 180.0
        else:
            scaling = self.circ_appear['circ_scaling']
        offset = -func(u_min) * scaling + angle_min * np.pi / 180.0
        self.func_f = lambda u: radius * np.cos(func(u) * scaling + offset)
        self.func_g = lambda u: radius * np.sin(func(u) * scaling + offset)
        if self.circ_appear['angle_tick_direction'] == 'inner':
            self.side = 'left'
        else:
            self.side = 'right'


class Circ_Block(object):
    """
    basic class for circle objects
    """

    def __init__(self, params={}):
        self.circ_scale_stack = []  # for circ scales
        self.params = {
            'inner_circle_radius': 2,
            'oblique_rotator': False,
            'arror_bar_width': 1.0,
            'rot_scale_extra_angle': 5.0 / 180.0 * np.pi,
            'rotator_fillet': 2.0
        }
        self.params.update(params)

    def _calculate_scaling_offset_(self, params):
        """
        calculates scalings for given params
        """
        func = lambda u: params['circ_sign'] * params['function'](u)
        u_min = params['u_min']
        u_max = params['u_max']
        angle_min = params['angle_min']
        angle_max = params['angle_max']
        scaling = (angle_max - angle_min) / (abs(func(u_max) - func(u_min))) * np.pi / 180.0
        if func(u_min) < func(u_max):
            offset = -func(u_min) * scaling + angle_min * np.pi / 180.0
        else:
            offset = -func(u_max) * scaling + angle_max * np.pi / 180.0
        return scaling, offset

    def _calculate_offset_(self, params):
        """
        assumes scaling is given
        """
        func = lambda u: params['circ_sign'] * params['function'](u)
        u_value = params['angle_offset_u_value']
        angle_offset = params['angle_offset_angle_value']
        scaling = params['circ_scaling']
        offset = -func(u_value) * scaling + angle_offset * np.pi / 180.0
        # offset=params['circ_sign']*(func(u_value)*scaling+angle_offset*math.pi/180.0)


        #        u_min=params['u_min']
        #        u_max=params['u_max']
        #        angle_min=params['angle_min']
        #        angle_max=params['angle_max']
        #        if func(u_min)<func(u_max):
        #            offset=-func(u_min)*scaling+angle_min*math.pi/180.0
        #        else:
        #            offset=-func(u_max)*scaling+angle_min*math.pi/180.0
        return offset

    def _draw_(self, params, f, g, canvas):
        """
        draws the circular scale
        """
        increase = params['circ_sign'] * params['function'](params['u_max']) - params['circ_sign'] * params['function'](
            params['u_min'])
        if params['angle_tick_direction'] == 'inner' and increase > 0:
            side = 'left'
        if params['angle_tick_direction'] == 'inner' and increase < 0:
            side = 'right'
        if params['angle_tick_direction'] == 'outer' and increase > 0:
            side = 'right'
        if params['angle_tick_direction'] == 'outer' and increase < 0:
            side = 'left'
        Nomo_Axis(func_f=f, func_g=g,
                  start=params['u_min'], stop=params['u_max'],
                  turn=-1, title=params['title'],
                  tick_levels=params['tick_levels'],
                  tick_text_levels=params['tick_text_levels'],
                  canvas=canvas, type=params['scale_type'],
                  manual_axis_data=params['manual_axis_data'],
                  side=side, axis_appear=params)

    def _draw_arrow_(self, F, G, angle, ccanvas):
        """
        draws an arrow, F amd G are constant functions
        """
        ccanvas.stroke(pyx.path.line(0.8 * F(0.0), 0.8 * G(0.0), 0.99 * F(0.0), 0.99 * G(0.0)),
                       [pyx.style.linewidth.thick, pyx.color.rgb.black,
                        pyx.deco.earrow([pyx.deco.stroked([pyx.color.rgb.black]),
                                     pyx.deco.filled([pyx.color.rgb.black])], size=0.3)])
        width = self.params['arror_bar_width']
        # ccanvas.stroke(self._bar_(angle,width,radius))

    def _draw_circle_(self, radius, ccanvas):
        """
        draws circle
        """
        ccanvas.stroke(pyx.path.circle(0, 0, radius), [pyx.style.linewidth.thin])
        """
        test:
        circ1=path.circle(0, 0, 1.5)
        circ2=path.circle(0, 1, 1.5)
        union_circ=self._union_paths_(circ1, circ2)
        ccanvas.stroke(union_circ, [style.linewidth.thin])
        ccanvas.stroke(union_circ, [deformer.smoothed(radius=0.5), color.rgb.red])
        """

    def _draw_pie_(self, start_angle, stop_angle, radius, canvas):
        """
        draws a pie
        """
        p1 = self._pie_(start_angle, stop_angle, radius)
        canvas.stroke(p1)

    def _pie_(self, start_angle, stop_angle, radius):
        """
        draws a pie
        """
        p = pyx.path.path(pyx.path.moveto(0, 0))
        if start_angle > stop_angle:
            start_angle, stop_angle = stop_angle, start_angle
        angle = start_angle
        while angle <= stop_angle:
            x = radius * np.cos(angle)
            y = radius * np.sin(angle)
            p.append(pyx.path.lineto(x, y))
            angle = angle + 0.002
        p.append(pyx.path.closepath())
        p1 = pyx.deformer.smoothed(self.params['rotator_fillet']).deform(p)
        return p1

    def _bar_(self, angle, width, radius):
        """
        draws a bar for arrow
        """
        # deformer.smoothed(1.0).deform(p)
        x = radius * np.cos(angle * np.pi / 180)
        y = radius * np.sin(angle * np.pi / 180)
        x1 = -self.params['inner_circle_radius'] / 2.0 * np.cos(angle * np.pi / 180)
        y1 = -self.params['inner_circle_radius'] / 2.0 * np.sin(angle * np.pi / 180)
        dx0 = np.cos((angle + 90) * np.pi / 180)
        dy0 = np.sin((angle + 90) * np.pi / 180)
        dx = -width / 2.0 * dx0 / np.sqrt(dx0 ** 2 + dy0 ** 2)
        dy = -width / 2.0 * dy0 / np.sqrt(dx0 ** 2 + dy0 ** 2)
        p = pyx.line(x1 + dx, y1 + dy, x + dx, y + dy)
        p.append(pyx.path.lineto(x - dx, y - dy))
        p.append(pyx.path.lineto(x1 - dx, y1 - dy))
        p.append(pyx.path.closepath())
        p1 = pyx.deformer.smoothed(width).deform(p)
        return p1

    def _draw_center_circle_(self, radius, ccanvas):
        """
        draws circle
        """
        ccanvas.stroke(pyx.path.circle(0, 0, radius), [pyx.style.linewidth.thin])
        ccanvas.stroke(pyx.path.line(-radius, 0, radius, 0), [pyx.style.linewidth.thin])
        ccanvas.stroke(pyx.path.line(0, -radius, 0, radius), [pyx.style.linewidth.thin])

    def _draw_title_ends_(self, params):
        """
        draws title to the ends
        """
        pass

    def _union_paths_(self, path1, path2):
        """
        makes union of two closed paths.
        """
        path1_i, path2_i = path1.intersect(path2)
        path1_i.sort()
        path2_i.sort()
        paths1 = path1.split(path1_i)
        paths2 = path2.split(path2_i)

        def longest(p):
            best = p[0]
            for pp in p:
                if pp.arclen() > best.arclen():
                    best = pp
            return best

        union = longest(paths1).joined(longest(paths2))
        union[-1].close()
        #        def longer(p1, p2):
        #            if p1.arclen() > p2.arclen():
        #                return p1
        #            return p2
        #        union = longer(*paths1) << longer(*paths2)
        # union=path1.split(path1_i)[-1]<<path2.split(path1_i)[-1]

        # union = path1 << path2
        # union[-1].close()
        return union


class Circ_Block_Type_1(Circ_Block):
    """
    type F1+F2+F3=0 circular slide rule
    scaling comes from F1
    offsets can be set for F2 and F3
    """

    def __init__(self):
        super(Circ_Block_Type_1, self).__init__()

    def set_block(self, block_params):
        """
        sets the block
        """
        self.f1_params = block_params['f1_params']
        self.f2_params = block_params['f2_params']
        self.f3_params = block_params['f3_params']
        self._check_initial_values_()
        self.block_params = block_params
        self.scaling, self.offset_f1 = self._calculate_scaling_offset_(self.f1_params)
        self.f1_params['circ_scaling'] = self.scaling
        self.f2_params['circ_scaling'] = self.scaling
        self.f3_params['circ_scaling'] = self.scaling
        self.offset_f2 = self._calculate_offset_(self.f2_params)
        self.offset_f3 = self._calculate_offset_(self.f3_params)
        self._calculate_funcs_()

    def _check_initial_values_(self):
        """
        parameters for the block
        """
        general_default = {
            'function': lambda u: u,
            'u_min': 1.0,
            'u_max': 10.0,
            'angle_min': 0.0,
            'angle_max': 180.0,
            'radius': 12,
            'angle_tick_direction': 'inner',  # or 'outer'
            'tick_levels': 4,
            'tick_text_levels': 2,
            'scale_type': 'linear',
            'text_distance_4': 1.0 / 4,
            'grid_length': 0.1,
            'grid_length_0': 0.5 / 4,
            'grid_length_1': 0.5 / 4,
            'grid_length_2': 0.4 / 4,
            'grid_length_3': 0.3 / 4,
            'grid_length_4': 0.2 / 4,
            'text_size': pyx.text.size.scriptsize,
            'text_size_0': pyx.text.size.tiny,
            'text_size_1': pyx.text.size.tiny,
            'text_size_2': pyx.text.size.tiny,
            'text_size_3': pyx.text.size.tiny,
            'text_size_4': pyx.text.size.tiny,
            'text_size_log_0': pyx.text.size.tiny,
            'text_size_log_1': pyx.text.size.tiny,
            'text_size_log_2': pyx.text.size.tiny,
            'text_size_manual': pyx.text.size.small,
            'text_distance_0': 1.0 / 4,
            'text_distance_1': 1.0 / 4,
            'text_distance_2': 1.0 / 4,
            'text_distance_3': 1.0 / 4,
            'text_distance_4': 1.0 / 4,
            'title_distance_center': 0.7,
            'title_opposite_tick': True,
            'title_draw_center': True,
            'text_format': "$%3.1f$",
            'full_angle': True,
            'extra_angle': 90.0,
            'text_horizontal_align_center': True,
            'manual_axis_data': {},
            'circ_scaling': None,
            'title': '',
            'angle_offset_u_value': 1,
            'angle_offset_angle_value': 1,
            'turn_relative': True}
        ##############################################################
        # F1
        params_default_f1 = general_default
        params_default_f1_0 = {
            'radius': 4,
            'angle_tick_direction': 'outer',
            'circ_sign': 1,
        }
        params_default_f1.update(params_default_f1_0)
        for key in params_default_f1:
            if not self.f1_params.has_key(key):
                self.f1_params[key] = params_default_f1[key]
            #        if self.f1_params['angle_tick_direction']=='inner':
            #            self.side_f1='left'
            #        else:
            #            self.side_f1='right'
        ##############################################################
        # F2
        params_default_f2 = general_default
        params_default_f2_0 = {
            'radius': 3.9,
            'angle_tick_direction': 'inner',
            'circ_sign': -1,
        }
        params_default_f2.update(params_default_f2_0)
        for key in params_default_f2:
            if not self.f2_params.has_key(key):
                self.f2_params[key] = params_default_f2[key]
            #        if self.f2_params['angle_tick_direction']=='inner':
            #            self.side_f2='left'
            #        else:
            #            self.side_f2='right'
        ##############################################################
        # F3
        params_default_f3 = general_default
        params_default_f3_0 = {
            'radius': 3,
            'angle_tick_direction': 'outer',
            'circ_sign': -1,
        }
        params_default_f3.update(params_default_f3_0)
        for key in params_default_f3:
            if not self.f3_params.has_key(key):
                self.f3_params[key] = params_default_f3[key]
            #        if self.f3_params['angle_tick_direction']=='inner':
            #            self.side_f3='left'
            #        else:
            #            self.side_f3='right'

    def _calculate_funcs_(self):
        """
        makes the real funcs
        """
        self.func_F1 = lambda u: self.f1_params['radius'] * np.cos(
            self.f1_params['circ_sign'] * self.f1_params['function'](u) * self.scaling + self.offset_f1)
        self.func_G1 = lambda u: self.f1_params['radius'] * np.sin(
            self.f1_params['circ_sign'] * self.f1_params['function'](u) * self.scaling + self.offset_f1)
        self.func_F2 = lambda u: self.f2_params['radius'] * np.cos(
            self.f2_params['circ_sign'] * self.f2_params['function'](u) * self.scaling + self.offset_f2)
        self.func_G2 = lambda u: self.f2_params['radius'] * np.sin(
            self.f2_params['circ_sign'] * self.f2_params['function'](u) * self.scaling + self.offset_f2)
        self.angle_2_min = self.f2_params['circ_sign'] * self.f2_params['function'](
            self.f2_params['u_min']) * self.scaling + self.offset_f2
        self.angle_2_max = self.f2_params['circ_sign'] * self.f2_params['function'](
            self.f2_params['u_max']) * self.scaling + self.offset_f2
        if self.angle_2_max < self.angle_2_min:
            self.angle_2_max, self.angle_2_min = self.angle_2_min, self.angle_2_max
        self.angle_2_min = self.angle_2_min - self.params['rot_scale_extra_angle']
        self.angle_2_max = self.angle_2_max + self.params['rot_scale_extra_angle']
        self.func_F3 = lambda u: self.f3_params['radius'] * np.cos(
            self.f3_params['circ_sign'] * self.f3_params['function'](u) * self.scaling + self.offset_f3)
        self.func_G3 = lambda u: self.f3_params['radius'] * np.sin(
            self.f3_params['circ_sign'] * self.f3_params['function'](u) * self.scaling + self.offset_f3)
        self.arrow_F = lambda u: (self.f3_params['radius'] - 0.05) * np.cos(u + self.offset_f2 + self.offset_f3)
        self.arrow_G = lambda u: (self.f3_params['radius'] - 0.25) * np.sin(u + self.offset_f2 + self.offset_f3)
        print("self.offset_f2 %g" % (self.offset_f2 * 180.0 / 3.1415))
        print("self.offset_f3 %g" % (self.offset_f3 * 180.0 / 3.1415))
        self.arrow_angle = (self.offset_f2 + self.offset_f3) * 180.0 / np.pi
        self.arrow_radius = self.f3_params['radius'] - 0.05

    def draw(self, ccanvas, rot_angle=0.0):
        """
        draws the scales
        """
        rotating_canvas = pyx.canvas.canvas()
        self.draw_oblique_rotator_edge(rotating_canvas)
        self._draw_(self.f1_params, self.func_F1, self.func_G1, ccanvas)
        self._draw_(self.f2_params, self.func_F2, self.func_G2, rotating_canvas)
        self._draw_(self.f3_params, self.func_F3, self.func_G3, ccanvas)
        self._draw_arrow_(self.arrow_F, self.arrow_G, self.arrow_angle,
                          rotating_canvas)
        # self._draw_pie_(self.angle_2_min,self.angle_2_max,self.f2_params['radius'],rotating_canvas)
        # self._draw_circle_(self.params['inner_circle_radius'],ccanvas)
        self._draw_circle_(self.f1_params['radius'] + 1, ccanvas)
        self._draw_center_circle_(0.2, ccanvas)
        # insert subcanvas into canvas
        ccanvas.insert(rotating_canvas, [pyx.trafo.rotate(rot_angle)])

    def draw_slide(self, ccanvas, rot_angle=0.0):
        """
        draws the rotating (transparent)
        """
        rotating_canvas = pyx.canvas.canvas()
        self.draw_oblique_rotator_edge(ccanvas)
        self._draw_(self.f2_params, self.func_F2, self.func_G2, rotating_canvas)
        self._draw_arrow_(self.arrow_F, self.arrow_G, self.arrow_angle,
                          rotating_canvas)

        # self._draw_pie_(self.angle_2_min,self.angle_2_max,self.f2_params['radius'],rotating_canvas)
        # self._draw_circle_(self.params['inner_circle_radius'],ccanvas)
        self._draw_circle_(self.f1_params['radius'] + 1, ccanvas)
        self._draw_center_circle_(0.2, ccanvas)
        # insert subcanvas into canvas
        ccanvas.insert(rotating_canvas, [pyx.trafo.rotate(rot_angle)])

    def draw_oblique_rotator_edge(self, ccanvas):
        """
        draws oblique rotating parts
        """
        # path around arrow
        bar_path = self._bar_(self.arrow_angle, self.params['arror_bar_width'],
                              self.arrow_radius)
        pie_path = self._pie_(self.angle_2_min, self.angle_2_max,
                              self.f2_params['radius'])
        circle_path = pyx.path.circle(0, 0, self.params['inner_circle_radius'])
        # ccanvas.stroke(pie_path,[deco.filled([color.cmyk.Gray])])
        # ccanvas.stroke(bar_path,[deco.filled([color.cmyk.Gray])])
        # ccanvas.stroke(circle_path,[deco.filled([color.cmyk.Gray])])
        p1 = self._union_paths_(circle_path, pie_path)
        p2 = self._union_paths_(bar_path, p1)
        p3 = pyx.deformer.smoothed(self.params['rotator_fillet'], obeycurv=0).deform(p2)
        ccanvas.fill(p3, [pyx.color.gray(0.95)])
        ccanvas.stroke(p3)
        self._draw_center_circle_(0.2, ccanvas)

    def draw_background(self, ccanvas, rot_angle=0.0):
        """
        draws the background
        """
        rotating_canvas = pyx.canvas.canvas()
        self._draw_(self.f1_params, self.func_F1, self.func_G1, ccanvas)
        self._draw_(self.f3_params, self.func_F3, self.func_G3, ccanvas)
        self._draw_circle_(self.f1_params['radius'] + 1, ccanvas)
        self._draw_center_circle_(0.2, ccanvas)


if __name__ == '__main__':
    c = pyx.canvas.canvas()
    appear = {
        'function': lambda u: u,
        'u_min': 1.0,
        'u_max': 12.0,
        'radius': 4,
        'angle_min': 75.0,
        'angle_max': 270.0,
        'scale_max': 10.0,
        'tick_levels': 4,
        'text_format': "$%3.1f$",
        'tick_text_levels': 2,
        'extra_angle': 90.0,
    }
    appear_1 = {
        'function': lambda u: u,
        'u_min': 1.0,
        'u_max': 9.0,
        'radius': 4,
        'angle_min': 0.0,
        'angle_max': 360.0,
        'scale_max': 10.0,
        'tick_levels': 5,
        'text_format': "$%3.1f$",
        'tick_text_levels': 2,
        'extra_angle': 90.0,
        'angle_tick_direction': 'outer',
    }
    circ_scale = Circ_Scale(canvas=c, circ_appear=appear)
    circ_scale.draw()
    circ_scale_1 = Circ_Scale(canvas=c, circ_appear=appear_1)
    circ_scale_1.draw()
    c.writePDFfile("test_circ")

    cc = pyx.canvas.canvas()
    cc_slide = pyx.canvas.canvas()
    cc_bg = pyx.canvas.canvas()
    para_1 = {'function': lambda u: 3 * u,
              'title': 'F1',
              'radius': 8,
              'u_min': 0.0,
              'u_max': 10.0,
              'angle_min': 0.0,
              'angle_max': 270.0}
    para_2 = {  # 'function':lambda u:10*math.log10(u),
                'function': lambda u: u,
                'title': 'F2',
                'radius': 7.95,
                'u_min': 1.0,
                'u_max': 10.0,
                'angle_offset_u_value': 1.0,
                'angle_offset_angle_value': 10.0,
                'scale_type': 'linear'}
    para_3 = {'function': lambda u: -u,
              'title': 'F3',
              'radius': 6,
              'u_min': -15.0,
              'u_max': 15.0,
              'angle_offset_u_value': 0.0,
              'angle_offset_angle_value': 180.0,
              'extra_angle': 0,
              'title_distance_center': 1.2,
              'text_horizontal_align_center': False}
    block_params = {'f1_params': para_1,
                    'f2_params': para_2,
                    'f3_params': para_3}

    circle_test = Circ_Block_Type_1()
    circle_test.set_block(block_params)
    circle_test.draw(cc, 45.0)
    circle_test.draw_slide(cc_slide, 0.0)
    circle_test.draw_background(cc_bg, 0.0)
    cc.writePDFfile("test_circ_combined.pdf")
    cc_slide.writePDFfile("test_circ_slide.pdf")
    cc_bg.writePDFfile("test_circ_bg.pdf")



#
#    This file is part of PyNomo -
#    a program to create nomographs with Python (http://pynomo.sourceforge.net/)
#
#    Copyright (C) 2007  Leif Roschier  <lefakkomies@users.sourceforge.net>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
This is example how to use lower level functions to build a nomograph.
Also Latex is used with oldstyle fonts that should be installed.
"""

from nomo_axis import *
from nomograph3 import *
import pyx

def f1(u):
    return 2 * (u * u - 1)


def g1(u):
    return 3 * u * (u + 1)


def h1(u):
    return -u * (u - 1)


def f3(v):
    return 2 * (2 * v + 1)


def g3(v):
    return 3 * (v + 1)


def h3(v):
    return -(v + 1) * (2 * v + 1)


def f2(w):
    return w


def g2(w):
    return 1


def h2(w):
    return -w * w


def f4(x):
    return 1


def g4(x):
    return 4 * math.log10(x)


pyx.text.set(mode="latex")
pyx.text.set(mode="latex", texmessagesdocclass=[pyx.text.texmessage.ignore])
nomograph = Nomograph3(f1=f1, f2=f2, f3=f3, g1=g1, g2=g2, g3=g3, h1=h1, h2=h2, h3=h3,
                       vk=[['u', 0.5, 'x', -5.0],
                           ['u', 0.5, 'y', 0.0],
                           ['u', 1.0, 'x', -5.0],
                           ['u', 1.0, 'y', 20.0],
                           ['w', 1.0, 'x', 5.0],
                           ['w', 1.0, 'y', 0.0],
                           ['w', 0.5, 'x', 5.0],
                           ['w', 0.5, 'y', 20.0]])
# nomograph._make_transformation_matrix_()
# \usepackage[T1]{fontenc}
# \usepackage[math]{anttor}
pyx.text.set(mode="latex")
# text.preamble(r"\usepackage[T1]{fontenc}")
# text.preamble(r"\usepackage[math]{anttor}")
# text.preamble(r"\oldstyle")

ccc = pyx.canvas.canvas()
gg3 = Nomo_Axis(func_f=nomograph.give_x3, func_g=nomograph.give_y3, start=1.0, stop=0.5,
                turn=-1, title='L', canvas=ccc, text_style='oldstyle')
gg1 = Nomo_Axis(func_f=nomograph.give_x1, func_g=nomograph.give_y1, start=0.5, stop=1.0,
                turn=-1, title='p', canvas=ccc, text_style='oldstyle')
gg1 = Nomo_Axis(func_f=nomograph.give_x2, func_g=nomograph.give_y2, start=1.0, stop=0.75,
                turn=1, title='h', canvas=ccc, text_style='oldstyle')
# gg2=Nomo_Axis(func_f=f4,func_g=g4,start=0.00001,stop=3,turn=1,title='log',canvas=ccc,type='log')
ccc.writePDFfile("nomotest_font5")
# SVN test 6



#    PyNomo - nomographs with Python
#    Copyright (C) 2007-2015  Leif Roschier
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" PyNomo - nomographs or nomograms with Python

Pynomo is a program to create (pdf) nomographs (nomograms)
using Python interpreter. A nomograph (nomogram) is a graphical
solution to an equation.
"""

__all__ = ["nomo_axis", "nomo_axis_func",
           "nomo_grid_box", "nomo_grid", "nomo_wrapper", "nomographer",
           "isopleth", "nomograph3"]



#
#    This file is part of PyNomo -
#    a program to create nomographs with Python (http://pynomo.sourceforge.net/)
#
#    Copyright (C) 2015  Leif Roschier  <lefakkomies@users.sourceforge.net>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .nomo_axis import Nomo_Axis
import math
import numpy as np
import pyx
import random
import sys


class Nomo_Grid:
    """
    class to make grids
    """

    def __init__(self, func_f, func_g, canvas, data={}):
        self.f = func_f
        self.g = func_g
        self.canvas = canvas
        data_default_values = {'u_start': 0.0,
                               'u_stop': 1.0,
                               'v_start': 0.0,
                               'v_stop': 1.0,
                               'u_values': [0.0, 1.0],
                               'v_values': [0.0, 1.0],
                               'v_texts_u_start': False,
                               'v_texts_u_stop': True,
                               'u_texts_v_start': False,
                               'u_texts_v_stop': True,
                               'u_line_color': pyx.color.rgb.black,
                               'v_line_color': pyx.color.rgb.black,
                               'u_text_color': pyx.color.rgb.black,
                               'v_text_color': pyx.color.rgb.black,
                               'u_line_width': pyx.style.linewidth.normal,
                               'v_line_width': pyx.style.linewidth.normal,
                               'text_distance': 0.25,
                               'circles': False,  # if marker circles
                               'text_prefix_u': '',  # for example r'$\alpha$='
                               'text_prefix_v': '',  # for example r'$\beta$='
                               'text_format_u': "$%4.4g$",
                               'text_format_v': "$%4.4g$",
                               'iterator_factor': 1.001,
                               }
        self.grid_data = data_default_values
        self.grid_data.update(data)
        # print self.grid_data
        self._draw_line_u_()
        self._draw_line_v_()

    def _draw_line_u_(self):
        """
        draws a math.single line from start to stop for variable u
        """
        start = self.grid_data['u_start']
        stop = self.grid_data['u_stop']
        # if text labels to start and/or stop
        start_texts = self.grid_data['u_texts_v_start']
        stop_texts = self.grid_data['u_texts_v_stop']
        line_color = self.grid_data['v_line_color']
        line_width = self.grid_data['v_line_width']
        for idx, v in enumerate(self.grid_data['v_values']):
            f_here, g_here = self._make_u_funcs_(v)
            if 'v_texts' not in self.grid_data:
                # print self.grid_data['text_prefix_v']
                self._draw_line_(f_here, g_here, start, stop,
                                 r"%s%s" % (self.grid_data['text_prefix_v'], self.grid_data['text_format_v']) % v,
                                 line_color,
                                 start_texts, stop_texts, self.grid_data['v_text_color'], line_width)
            else:
                self._draw_line_(f_here, g_here, start, stop,
                                 self.grid_data['v_texts'][idx], line_color,
                                 start_texts, stop_texts, self.grid_data['v_text_color'], line_width)
                # print "v=%f"%v

    def _draw_line_v_(self):
        """
        draws a math.single line from start to stop for variable v
        """
        start = self.grid_data['v_start']
        stop = self.grid_data['v_stop']
        # if text labels to start and/or stop
        start_texts = self.grid_data['v_texts_u_start']
        stop_texts = self.grid_data['v_texts_u_stop']
        line_color = self.grid_data['u_line_color']
        line_width = self.grid_data['u_line_width']
        for idx, u in enumerate(self.grid_data['u_values']):
            f_here, g_here = self._make_v_funcs_(u)
            if 'u_texts' not in self.grid_data:
                self._draw_line_(f_here, g_here, start, stop,
                                 r"%s%s" % (self.grid_data['text_prefix_u'], self.grid_data['text_format_u']) % u,
                                 line_color,
                                 start_texts, stop_texts, self.grid_data['u_text_color'], line_width)
            else:
                self._draw_line_(f_here, g_here, start, stop,
                                 self.grid_data['u_texts'][idx], line_color,
                                 start_texts, stop_texts, self.grid_data['u_text_color'], line_width)
                # print "u=%f"%u

    def _make_u_funcs_(self, v_value):
        """
        copied trick to solve function defitions inside loop
        (I could not figure out how to use lambda...)
        """

        def f(u): return self.f(u, v_value)

        def g(u): return self.g(u, v_value)

        return f, g

    def _make_v_funcs_(self, u_value):
        """
        copied trick to solve function defitions inside loop
        (I could not figure out how to use lambda...)
        """

        def f(v): return self.f(u_value, v)

        def g(v): return self.g(u_value, v)

        return f, g

    def _draw_line_(self, f, g, start, stop, title, axis_color=pyx.color.rgb.red,
                    start_texts=False, stop_texts=True, text_color=pyx.color.rgb.black,
                    line_width=pyx.style.linewidth.normal):
        if start > 0 and stop > 0:
            du = (max(start, stop) - min(start, stop)) * 1e-10  # was 1e-10
            # print "start:%g stop:%g"%(start,stop)
            # print "testing du = %g"%du
        else:
            du = np.fabs(start - stop) * 1e-5
        # approximate line length is found
        line_length_straigth = np.sqrt((f(start) - f(stop)) ** 2 + (g(start) - g(stop)) ** 2)
        random.seed(0.0)  # so that mistakes always the same
        for dummy in range(100):
            first = random.uniform(start, stop)
            second = random.uniform(start, stop)
            temp = np.sqrt((f(first) - f(second)) ** 2 + (g(first) - g(second)) ** 2)
            if temp > line_length_straigth:
                line_length_straigth = temp
                # print "length: %f"%line_length_straigth
        sections = 350.0  # about number of sections
        section_length = line_length_straigth / sections
        line = pyx.path.path(pyx.path.moveto(f(start), g(start)))
        line.append(pyx.path.lineto(f(start), g(start)))
        u = start
        laskuri = 1
        up_factor = self.grid_data['iterator_factor']
        down_factor = 1.0 / up_factor
        delta_u = du  # initial
        while True:
            if (u + delta_u) < stop:
                dx = (f(u + du) - f(u))
                dy = (g(u + du) - g(u))
                dl = np.sqrt(dx ** 2 + dy ** 2)
                delta_u = du * section_length / dl
                # let's calculate actual length
                # and iterate until length is in factor 2 from target
                while True:
                    # u_delta_step = min(stop,u+delta_u) # in order to avoid going over range
                    if (u + delta_u) >= stop:  # stop if out of range
                        break
                    delta_x = f(u + delta_u) - f(u)
                    delta_y = g(u + delta_u) - g(u)
                    delta_l = np.sqrt(delta_x ** 2 + delta_y ** 2)
                    if delta_l > 5.0 * section_length:
                        delta_u = delta_u * down_factor  # 0.999
                        # print "delta_u math.pienenee:%f"%delta_u
                    else:
                        if delta_l < section_length / 5.0:
                            delta_u = delta_u * up_factor  # 1.001
                            # print "delta_u kasvaa:%f"%delta_u
                            # print "u+delta_u:%g, stop:%g"%(u+delta_u,stop)
                    if delta_l <= 5.0 * section_length and delta_l >= 0.2 * section_length:
                        # print "u:%g:x=%g, y=%g"%(u,f(u),g(u))
                        break

                u += delta_u
                # print u,stop
                laskuri = laskuri + 1
                line.append(pyx.path.lineto(f(u), g(u)))
            else:
                line.append(pyx.path.lineto(f(stop), g(stop)))
                # print laskuri
                sys.stdout.write('.')
                break

        self.canvas.stroke(line, [line_width, axis_color])
        # start number
        if start_texts:  # set texts to to start
            self._set_text_to_grid_(f, g, start, du, title, axis_color, text_color)
        if stop_texts:  # set texts to stop
            self._set_text_to_grid_(f, g, stop, -du, title, axis_color, text_color)
        if self.grid_data['circles']:
            self.canvas.fill(pyx.path.circle(f(start), g(start), 0.03), [axis_color])
            self.canvas.fill(pyx.path.circle(f(stop), g(stop), 0.03), [axis_color])
            # print "line drawn"

    def _set_text_to_grid_(self, f, g, u, du, title, axis_color, text_color=pyx.color.rgb.black):
        """
        draws text to the end of gridline
        """
        dx = (f(u + du) - f(u))
        dy = (g(u + du) - g(u))
        if np.sqrt(dx ** 2 + dy ** 2) == 0:
            dx_unit = 0
            dy_unit = 0
        else:
            dx_unit = dx / np.sqrt(dx ** 2 + dy ** 2)
            dy_unit = dy / np.sqrt(dx ** 2 + dy ** 2)
        if dy_unit != 0:
            """
            print "du %g"%du
            print "dx %g"%dx
            print "dy %g"%dy
            print "dx_unit %g"%dx_unit
            print "dy_unit %g"%dy_unit
            """
            angle = -math.atan(dx_unit / dy_unit) * 180 / math.pi
        else:
            angle = 0
        text_distance = 0.5
        # if dy<=0:
        if dy >= 0.0:
            if (angle - 90.0) <= -90.0:
                angle = angle + 180.0
            if dx_unit > 0.0:
                text_attr = [pyx.text.valign.middle, pyx.text.halign.right, pyx.text.size.small,
                             pyx.trafo.rotate(angle - 90)]
            if dx_unit <= 0.0:
                text_attr = [pyx.text.valign.middle, pyx.text.halign.left, pyx.text.size.small,
                             pyx.trafo.rotate(angle - 90)]
        else:
            if (angle + 90.0) >= 90.0:
                angle = angle - 180.0
            if dx_unit > 0.0:
                text_attr = [pyx.text.valign.middle, pyx.text.halign.right, pyx.text.size.small,
                             pyx.trafo.rotate(angle + 90)]
            if dx_unit <= 0.0:
                text_attr = [pyx.text.valign.middle, pyx.text.halign.left, pyx.text.size.small,
                             pyx.trafo.rotate(angle + 90)]
        """
        copied from nomo_axis.py
        if dy_units[idx]<0:
            text_attr=[pyx.text.valign.middle,pyx.text.halign.right,text_size,pyx.trafo.rotate(angles[idx])]
        else:
            text_attr=[pyx.text.valign.middle,pyx.text.halign.left,text_size,pyx.trafo.rotate(angles[idx])]
        text_list.append((self._put_text_(u),f(u)+text_distance*dy_units[idx],
                          g(u)-text_distance*dx_units[idx],text_attr))
        """
        text_distance = self.grid_data['text_distance']
        text_attr = text_attr + [text_color]
        self.canvas.text(f(u) - text_distance * dx_unit,
                         g(u) - text_distance * dy_unit,
                         title, text_attr)
        # self.canvas.fill(pyx.path.circle(f(u), g(u), 0.03),[axis_color])


if __name__ == '__main__':
    # functions for solartime
    # taken from solareqns.pdf from
    # http://www.srrb.noaa.gov/highlights/sunrise/solareqns.PDF
    def gamma(day):
        return 2 * math.pi / 365.0 * (day - 1 + 0.5)


    def eq_time(day):
        gamma0 = gamma(day)
        return 229.18 * (0.000075 + 0.001868 * np.cos(gamma0) - 0.032077 * np.sin(gamma0)
                         - 0.014615 * np.cos(2 * gamma0) - 0.040849 * np.sin(2 * gamma0))


    # np.mean correction, with constant correction we make less than 1.5 minutes error
    # in time axis
    temp_a = np.arange(0, 2 * math.pi, 0.001)
    temp_b = eq_time(temp_a)
    correction = np.mean(temp_b)  # this is about four minutes


    def eq_declination(day):
        g0 = gamma(day)
        return 0.006918 - 0.399912 * np.cos(g0) + 0.070257 * math.sin(g0) - 0.006758 * np.cos(2 * g0) \
               + 0.000907 * math.sin(2 * g0) - 0.002697 * np.cos(3 * g0) + 0.00148 * math.sin(3 * g0)


    def tst(day, hour):
        return hour * 60.0 + eq_time(day)


    def ha(day, hour):
        return tst(day, hour) / 4.0 - 180.0


    multiplier_x = 20.0
    multiplier_y = 10.0


    def f(lat, day):
        dec = eq_declination(day)
        return multiplier_x * (np.cos(lat * math.pi / 180.0) * np.cos(dec)) / (
                    1.0 + (np.cos(lat * math.pi / 180.0) * np.cos(dec)))


    def g(lat, day):
        dec = eq_declination(day)  # in radians
        return multiplier_y * (np.sin(lat * math.pi / 180.0) * math.sin(dec)) / (
                    1.0 + (np.cos(lat * math.pi / 180.0) * np.cos(dec)))


    def f1(dummy):
        return 0.0


    def g1(fii):
        return multiplier_y * np.cos(fii * math.pi / 180.0)


    def f3(dummy):
        return multiplier_x


    def g3(h):
        hr = (h * 60.0 + correction) / 4.0 - 180.0
        return -multiplier_y * np.cos(hr * math.pi / 180.0)


    days_in_month = (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)
    times1 = []
    for idx in range(0, 12):
        times1.append(sum(days_in_month[0:idx]) + 1)

    # times=linspace(0,350,10)
    times = np.arange(0.0, 360.0, 10.0, dtype=np.double).tolist()
    time_titles = ['January', 'February', 'March', 'April', 'May', 'June',
                   'July', 'August', 'September', 'October', 'November', 'December']
    # times.append(365)
    data = {'u_start': 20.0,  # latitude
            'u_stop': 80.0,
            'v_start': times1[0],  # day
            'v_stop': times1[-1],
            'u_values': [20.0, 30.0, 40.0, 50.0, 60.0, 70.0, 80.0],
            # 'v_values':[0.0,60.0,120.0,180.0,240.0,300.0,365.0]
            'v_values': times1,
            'v_texts': time_titles}

    c = pyx.canvas.canvas()
    gridi = Nomo_Grid(f, g, c, data=data)
    ax1 = Nomo_Axis(func_f=f1, func_g=g1,
                    start=0.0, stop=90.0,
                    title=r'Solar zenith angle $\phi$', canvas=c, type='linear',
                    turn=-1,
                    tick_levels=2, tick_text_levels=1,
                    side='left')

    ax2 = Nomo_Axis(func_f=f3, func_g=g3,
                    start=0.0, stop=23.0,
                    title=r'Hour (h)', canvas=c, type='linear',
                    turn=-1,
                    tick_levels=3, tick_text_levels=1,
                    side='right')

    c.writePDFfile("test_nomo_grid")
    # let's get
    test_h = 10.0
    test_day = 70.0
    test_lat = 60.0
    test_ha = test_h * 60.0 / 4.0 - 180.0
    test_dec = eq_declination(test_day)
    test_cos_phi = math.sin(test_lat * math.pi / 180.0) * math.sin(test_dec) + \
                   np.cos(test_lat * math.pi / 180.0) * np.cos(test_dec) * np.cos(test_ha * math.pi / 180.0)
    # print acos(test_cos_phi)*180/pi



# -*- encoding: utf-8 -*-
#
#    PyNomo - nomographs with Python
#    Copyright (C) 2007- 2023  Leif Roschier
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
""" PyNomo - nomographs or nomograms with Python

Pynomo is a program to create (pdf) nomographs (nomograms)
using Python interpreter. A nomograph (nomogram) is a graphical
solution to an equation.
"""
# example from: https://github.com/pypa/sampleproject/blob/master/setup.py
from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='PyNomo',
      version='0.3.6',
      description='PyNomo - Python Nomograms',
      long_description=long_description,
      author='Leif Roschier',
      author_email='leif.roschier@iki.fi',
      url='http://pynomo.org/',
      download_url='https://github.com/lefakkomies/pynomo',
      packages=['pynomo'],
      license='GPL',
      platforms='OS Independent',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Topic :: Scientific/Engineering :: Visualization',
          'Topic :: Multimedia :: Graphics',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'License :: OSI Approved :: GNU General Public License (GPL)',
          'Programming Language :: Python',
      ],
      keywords='nomograph nomogram graphics calculator visualization',
      # install numpy, scipy, pyx and latex manually :(
      #install_requires=['numpy', 'scipy'],  # have to install pyx and tex/latex manually
      )



# -*- encoding: utf-8 -*-
#
#    This file is part of PyNomo -
#    a program to create nomographs with Python (https://github.com/lefakkomies/pynomo)
#
#    Copyright (C) 2007-2019  Leif Roschier  <lefakkomies@users.sourceforge.net>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import math
import pyx
import copy, re
#from scipy.optimize import *
import scipy.optimize
from numpy import arange
import warnings


class Isopleth_Wrapper(object):
    """
    class to hold all isopleths and control everything related to them
    """

    def __init__(self, params):
        self.isopleth_list = []  # list of isopleth objects
        self.solutions = []  # list of dictionaries (solutions)
        self.ref_tag_number = 1  # to separate type 3 ref scales
        self.nomographer_params = params['isopleth_params']  # main params

    def add_isopleth_block(self, block, block_para):
        """
        Add block of type derived from Isopleth_Block
        """
        # type 1
        if block_para['block_type'] == 'type_1':
            iso_block = Isopleth_Block_Type_1(block.atom_stack, block_para)
            self.isopleth_list.append(iso_block)
        # type 2
        if block_para['block_type'] == 'type_2':
            iso_block = Isopleth_Block_Type_2(block.atom_stack, block_para)
            self.isopleth_list.append(iso_block)
        # type 7
        if block_para['block_type'] == 'type_7':
            iso_block = Isopleth_Block_Type_7(block.atom_stack, block_para)
            self.isopleth_list.append(iso_block)
        # type 8
        if block_para['block_type'] == 'type_8':
            iso_block = Isopleth_Block_Type_8(block.atom_stack, block_para)
            self.isopleth_list.append(iso_block)
        # type 9
        if block_para['block_type'] == 'type_9':
            iso_block = Isopleth_Block_Type_9(block.atom_stack, block_para)
            self.isopleth_list.append(iso_block)
        # type 10
        if block_para['block_type'] == 'type_10':
            iso_block = Isopleth_Block_Type_10(block.atom_stack, block_para)
            self.isopleth_list.append(iso_block)
        # type 6
        if block_para['block_type'] == 'type_6':
            iso_block = Isopleth_Block_Type_6(block.atom_stack, block_para)
            self.isopleth_list.append(iso_block)
        # type 3
        if block_para['block_type'] == 'type_3':
            # handle type 3 as multiple type 1
            N = block.N
            ref_N = N - 3
            ref_atoms = block.atom_stack[N:(N + ref_N)]
            # set ref axes
            for idx, ref_atom in enumerate(ref_atoms):
                ref_atom.params['tag'] = 'ref' + repr(idx) + repr(self.ref_tag_number)
            self.ref_tag_number += 1
            atoms = block.atom_stack[:N]
            atom_stack_start = [atoms[0], atoms[1], ref_atoms[0]]
            atom_stack_stop = [ref_atoms[-1], atoms[-2], atoms[-1]]
            center_atom_stack = []
            for idx in range(2, N - 2):
                center_atom_stack.append([ref_atoms[idx - 2], atoms[idx], ref_atoms[idx - 1]])
            # make block params
            # start
            block_para_start = copy.deepcopy(block_para)
            block_para_start['isopleth_values'] = []
            for idx, isopleth_values in enumerate(block_para['isopleth_values']):
                block_para_start['isopleth_values'].append([block_para['isopleth_values'][idx][0],
                                                            block_para['isopleth_values'][idx][1], 'x'])
            # stop
            block_para_stop = copy.deepcopy(block_para)
            block_para_stop['isopleth_values'] = []
            for idx, isopleth_values in enumerate(block_para['isopleth_values']):
                block_para_stop['isopleth_values'].append(['x', block_para['isopleth_values'][idx][-2],
                                                           block_para['isopleth_values'][idx][-1]])
            # middle
            block_para_middles = []
            for idx in range(3, N - 1):
                block_para_middles.append(copy.deepcopy(block_para))
                block_para_middles[-1]['isopleth_values'] = []
                for idx1, isopleth_values in enumerate(block_para['isopleth_values']):
                    block_para_middles[-1]['isopleth_values'].append(
                        ['x', block_para['isopleth_values'][idx1][idx - 1], 'x'])
            # do the list
            self.isopleth_list.append(Isopleth_Block_Type_1(atom_stack_start, block_para_start))
            for idx, atom_stack in enumerate(center_atom_stack):
                self.isopleth_list.append(Isopleth_Block_Type_1(atom_stack, block_para_middles[idx]))
            self.isopleth_list.append(Isopleth_Block_Type_1(atom_stack_stop, block_para_stop))

        # type 4
        if block_para['block_type'] == 'type_4':
            atoms = block.atom_stack
            atom_stack_12 = [atoms[0], atoms[1], atoms[4]]  # 4 = ref line
            atom_stack_34 = [atoms[2], atoms[3], atoms[4]]
            # tag the reference line
            atoms[4].params['tag'] = 'ref_type4' + repr(self.ref_tag_number)
            self.ref_tag_number += 1
            # make blocks
            block_para_12 = copy.deepcopy(block_para)
            block_para_12['isopleth_values'] = []
            for idx, isopleth_values in enumerate(block_para['isopleth_values']):
                block_para_12['isopleth_values'].append([block_para['isopleth_values'][idx][0],
                                                         block_para['isopleth_values'][idx][1], 'x'])
            block_para_34 = copy.deepcopy(block_para)
            block_para_34['isopleth_values'] = []
            for idx, isopleth_values in enumerate(block_para['isopleth_values']):
                block_para_34['isopleth_values'].append([block_para['isopleth_values'][idx][2],
                                                         block_para['isopleth_values'][idx][3], 'x'])
            self.isopleth_list.append(Isopleth_Block_Type_1(atom_stack_12, block_para_12))
            self.isopleth_list.append(Isopleth_Block_Type_1(atom_stack_34, block_para_34))

        # type 5 (contour)
        if block_para['block_type'] == 'type_5':
            iso_block = Isopleth_Block_Type_5(block.atom_stack, block_para, block)
            self.isopleth_list.append(iso_block)

    def draw(self, canvas):
        """
        solves isopleths and draws them
        """
        for isopleth in self.isopleth_list:
            isopleth.calc_atoms()
        self._solve_()
        for idx, isopleth in enumerate(self.isopleth_list):
            p = self.nomographer_params
            isopleth.draw(canvas, p)

    def _solve_(self):
        """
        solves unknown values
        """
        solutions_updated = True
        while solutions_updated:
            for idx, isopleth in enumerate(self.isopleth_list):
                isopleth.solve(self.solutions)
                # take initial values (they are most correct)
                isopleth.find_initial_solutions(self.solutions)
            # updates solutions
            solutions_updated = False
            for idx, isopleth in enumerate(self.isopleth_list):
                update = isopleth.update_solutions(self.solutions)
                if update:
                    solutions_updated = True
        # check for error
        for idx, isopleth in enumerate(self.isopleth_list):
            isopleth.check_if_all_solutions_found(self.solutions)


class Isopleth_Block(object):
    """
    parent class for isopleth generation for blocks.
    Isopleths should be instanced once Atoms have final transformations
    """

    def __init__(self, atom_stack, params):
        """
        params is for example:
        {
        'isopleth_values':[['x',0.1,0.2]],
        }
        if x is found outside this implementation, 'x' is replaced
        by tuple (x,y) of the coordinate pair
        """
        self.params = params
        # self.block=block
        self.isopleth_values = params['isopleth_values']
        self.atom_stack = atom_stack
        self.draw_coordinates = []  # coordinates [[x1,y1,x2,y2,x3,y3],...] to be drawn
        self.other_points = []  # list of list of additional solution coordinates

    def calc_atoms(self):
        """
        calculates coordinates for atoms
        """
        for atom in self.atom_stack:
            # calculates lines (list of coordinates)
            atom.calc_line_and_sections()

        #    def _replace_found_values_(self,found_dict={}):
        #        """
        #        replaces found variables with the coordinates
        #        found_dict is of form (for example)
        #        {
        #        'x1':(1.0,2.0),
        #        'y2':(3.4,2.1)
        #        }
        #        """
        #        for key in found_dict.keys():
        #            if self.params['points'].count(key)>0:
        #                idx=self.params['points'].index(key)
        #                self.params['points'][idx]=found_dict[key]

    @staticmethod
    def _calc_distance_(x0, y0, x1, y1, x2, y2):
        """
        Calculates distance of point (x0,y0) from line passing
        through points (x1,y1), (x2,y2)
        """
        return abs((x2 - x1) * (y1 - y0) - (x1 - x0) * (y2 - y1)) / math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)

    @staticmethod
    def _calc_distance_points_(x1, y1, x2, y2):
        """
        calcs distance between two points
        """
        return math.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2)

    def _find_closest_point_old_(self, line, x1, y1, x2, y2):
        """
        finds closest point(S) of isopleth and axis (scale)
        """
        x = line[0][0]
        y = line[0][1]
        distances = []
        smallest_distance = self._calc_distance_(x, y, x1, y1, x2, y2)
        distances.append(smallest_distance)
        smallest_idx = 0
        for idx, (x, y) in enumerate(line):
            distance = self._calc_distance_(x, y, x1, y1, x2, y2)
            distances.append(distance)
            if distance < smallest_distance:
                smallest_distance = distance
                smallest_idx = idx
        if smallest_idx == 0:
            idx2 = 1
        if smallest_idx == len(line):
            idx2 = len(line) - 1
        if 0 < smallest_idx < len(line):
            if distances[smallest_idx - 1] < distances[smallest_idx + 1]:
                idx2 = smallest_idx - 1
            else:
                idx2 = smallest_idx + 1

            #        sum_distance=distances[smallest_idx]+distances[idx2]
            #        middle_x=distances[smallest_idx]/sum_distance*line[smallest_idx][0]+\
            #                 distances[idx2]/sum_distance*line[idx2][0]
            #        middle_y=distances[smallest_idx]/sum_distance*line[smallest_idx][1]+\
            #                 distances[idx2]/sum_distance*line[idx2][1]
        # a better
        middle_x, middle_y = self._two_line_intersection_(line[smallest_idx][0], line[smallest_idx][1],
                                                          line[idx2][0], line[idx2][1],
                                                          x1, y1, x2, y2)
        return middle_x, middle_y

    def _find_closest_point_(self, sections, x1, y1, x2, y2):
        """
        finds closest point(S) of isopleth and axis (scale)
        """
        f1 = 1.0 - 1e-12
        f2 = 1.0 + 1e-12
        interps = []
        for idx, (x1s, y1s, x2s, y2s) in enumerate(sections):
            x_inter, y_inter = self._two_line_intersection_(x1s, y1s, x2s, y2s, x1, y1, x2, y2)
            # check if instersection
            # if (f1*min(x1s,x2s)<=x_inter<=f2*max(x1s,x2s)) and (f1*min(y1s,y2s)<=y_inter<=f2*max(y1s,y2s)):
            if self._between_(x1s, y1s, x2s, y2s, x_inter, y_inter):
                interps.append((x_inter, y_inter))
        if len(interps) < 1:
            interps.append((-10, -10))  # dummy point
        return interps[0][0], interps[0][1]

    @staticmethod
    def _between_(x1s, y1s, x2s, y2s, x, y):
        """
        checks if (x,y) in rectangle of (x1s,y1s)-(x2s,y2s)
        """
        f1 = 1.0 - 1e-12
        f2 = 1.0 + 1e-12
        f3 = 1e-6
        xs_min = min(x1s, x2s)
        if xs_min > 0:
            xs_min *= f1
        else:
            xs_min *= f2
        xs_max = max(x1s, x2s)
        if xs_max > 0:
            xs_max *= f2
        else:
            xs_max *= f1
        ys_min = min(y1s, y2s)
        if ys_min > 0:
            ys_min *= f1
        else:
            ys_min *= f2
        ys_max = max(y1s, y2s)
        if ys_max > 0:
            ys_max *= f2
        else:
            ys_max *= f1
        # trick to make little little over zero
        if xs_min == 0:
            xs_min = -xs_max * f3
        if xs_max == 0:
            xs_max = -xs_min * f3
        if ys_min == 0:
            ys_min = -ys_max * f3
        if ys_max == 0:
            ys_max = -ys_min * f3
        if xs_min <= x <= xs_max and ys_min <= y <= ys_max:
            return True
        else:
            return False

    def _find_closest_other_points_(self, sections, x1, y1, x2, y2, x_found, y_found):
        """
        finds closest point(S) of isopleth and axis (scale)
        x_found, y_found are found already before
        """
        f = 1 - 1e-12  # factor to reduce double hits
        interps = []
        for idx, (x1s, y1s, x2s, y2s) in enumerate(sections):
            x_inter, y_inter = self._two_line_intersection_(x1s, y1s, x2s * f, y2s * f, x1, y1, x2, y2)
            # check if instersection
            if (min(x1s, x2s * f) <= x_inter <= max(x1s, x2 * f)) and (
                    min(y1s, y2s * f) <= y_inter <= max(y1s, y2s * f)):
                if not x_found == x_inter and not y_found == y_inter:
                    interps.append((x_inter, y_inter))
        return interps

    @staticmethod
    def collinear(x1, y1, x2, y2, x3, y3):
        determinant = x1 * (y2 - y3) + x2 * (y3 - y1) + x3 * (y1 - y2)
        if abs(determinant) < 1e-3:
            return True
        else:
            return False

    def find_farthest_pair_extra(self, x1, y1, x2, y2, x3, y3, idx):
        """
        finds farthest pair including extra points
        """
        xf1, yf1, xf2, yf2 = self.find_farthest_pair(x1, y1, x2, y2, x3, y3)
        for points in self.other_points[idx]:
            for (x, y) in points:
                xf1, yf1, xf2, yf2 = self.find_farthest_pair(xf1, yf1, xf2, yf2, x, y)
        return xf1, yf1, xf2, yf2

    def find_farthest_pair(self, x1, y1, x2, y2, x3, y3):
        """
        finds farthest two points of three
        """
        dist_12 = self._calc_distance_points_(x1, y1, x2, y2)
        dist_13 = self._calc_distance_points_(x1, y1, x3, y3)
        dist_23 = self._calc_distance_points_(x2, y2, x3, y3)
        if dist_12 >= dist_13 and dist_12 >= dist_23:
            return x1, y1, x2, y2
        if dist_13 >= dist_23 and dist_13 >= dist_12:
            return x1, y1, x3, y3
        if dist_23 >= dist_13 and dist_23 >= dist_12:
            return x2, y2, x3, y3

    def draw(self, canvas, draw_params=[{}]):
        """
        draws the isopleth
        """
        for idx, (x1, y1, x2, y2, x3, y3) in enumerate(self.draw_coordinates):
            xx1, yy1, xx2, yy2 = self.find_farthest_pair_extra(x1, y1, x2, y2, x3, y3, idx)
            # print xx1,yy1,xx2,yy2
            # check for collinearity
            #            if not self.collinear(x1, y1, x2, y2, x3, y3):
            #                print "found points not collinear in isopleth..."
            if len(draw_params) > idx:
                p = draw_params[idx]
            else:
                p = draw_params[len(draw_params) - 1]
            draw_params_list = self.parse_isopleth_params(p)
            color_param = draw_params_list[0]
            #            canvas.stroke(pyx.path.line(xx1,yy1,xx2,yy2),[pyx.color.cmyk.Black,
            #                                                    pyx.style.linewidth.thick,
            #                                                    pyx.style.linestyle.dashed])
            canvas.stroke(pyx.path.line(xx1, yy1, xx2, yy2), draw_params_list)
            circle_radius = self.parse_circle_size(p)
            self._draw_circle_(canvas, x1, y1, circle_radius, color_param)
            self._draw_circle_(canvas, x2, y2, circle_radius, color_param)
            self._draw_circle_(canvas, x3, y3, circle_radius, color_param)
        for idx, line_points in enumerate(self.other_points):
            if len(draw_params) > idx:
                p = draw_params[idx]
            else:
                p = draw_params[len(draw_params) - 1]
            circle_radius = self.parse_circle_size(p)
            draw_params_list = self.parse_isopleth_params(p)
            color_param = draw_params_list[0]
            for points in line_points:
                for (x, y) in points:
                    self._draw_circle_(canvas, x, y, circle_radius, color_param)

    def _draw_circle_(self, canvas, x, y, r, circle_color=pyx.color.cmyk.Black):
        """
        draws marker circle
        """
        canvas.fill(pyx.path.circle(x, y, r), [pyx.color.rgb.white])
        canvas.stroke(pyx.path.circle(x, y, r), [circle_color])

    def solve(self, solutions):
        """
        parent class to be overriden, solves coordinates
        """
        pass

    def check_if_all_solutions_found(self, solutions):
        """
        this has right now no function
        """
        all_found = True
        """
        for atom_idx,atom in enumerate(self.atom_stack):
            for idx,dummy in enumerate(solutions):
                if not isinstance(self.isopleth_values[idx][atom_idx],(int,float,tuple)):
                    pass
                    #print "not all isopleths solved yet..."
                    """

    def find_initial_solutions(self, solutions):
        """
        Finds initial solutions
        """
        for atom_idx, atom in enumerate(self.atom_stack):
            if not atom.params['tag'] == 'none':
                for idx, dummy in enumerate(solutions):
                    # store only true (x,y) tuples
                    if isinstance(self.isopleth_values[idx][atom_idx], (tuple)):
                        solutions[idx][atom.params['tag']] = self.isopleth_values[idx][atom_idx]
                    if isinstance(self.isopleth_values[idx][atom_idx], (int, float)):
                        value = self.isopleth_values[idx][atom_idx]
                        x = atom.give_x(value) - atom.params['align_x_offset']
                        y = atom.give_y(value) - atom.params['align_y_offset']
                        solutions[idx][atom.params['tag']] = (x, y)
            else:
                if not atom.params['dtag'] == 'none':
                    for idx, dummy in enumerate(solutions):
                        # store only true (x,y) tuples
                        if isinstance(self.isopleth_values[idx][atom_idx], (tuple)):
                            solutions[idx][atom.params['dtag']] = self.isopleth_values[idx][atom_idx]
                        if isinstance(self.isopleth_values[idx][atom_idx], (int, float)):
                            value = self.isopleth_values[idx][atom_idx]
                            x = atom.give_x(value) - atom.params['align_x_offset']
                            y = atom.give_y(value) - atom.params['align_y_offset']
                            solutions[idx][atom.params['dtag']] = (x, y)

    def update_solutions(self, solutions):
        """
        Updates solutions
        """
        solutions_updated = False
        for idx, solution in enumerate(solutions):
            for key in solution.keys():
                for atom_idx, atom in enumerate(self.atom_stack):
                    if atom.params['tag'] == key or atom.params['dtag'] == key:
                        if isinstance(self.isopleth_values[idx][atom_idx], str):
                            self.isopleth_values[idx][atom_idx] = solution[key]
                            solutions_updated = True
        return solutions_updated

    def _check_if_enough_params_(self, idx):
        """
        parent class to be overriden, checks if enough params to solve
        """
        pass

    def _two_line_intersection_(self, x1, y1, x2, y2, x3, y3, x4, y4):
        """
        intersection of lines (x1,y1)-(x2,y2) and (x3,y3)-(x4,y4)
        """
        # here we catch divide-by-zero and set value to inf
        with warnings.catch_warnings():
            warnings.filterwarnings('error')
            try:
                x = self._det_(self._det_(x1, y1, x2, y2), (x1 - x2), self._det_(x3, y3, x4, y4), (x3 - x4)) / \
                    self._det_(x1 - x2, y1 - y2, x3 - x4, y3 - y4)
            except:
                # print("ValueError on L488 %g" % (self._det_(x1 - x2, y1 - y2, x3 - x4, y3 - y4)))
                x = float('Inf')
            try:
                y = self._det_(self._det_(x1, y1, x2, y2), (y1 - y2), self._det_(x3, y3, x4, y4), (y3 - y4)) / \
                    self._det_(x1 - x2, y1 - y2, x3 - x4, y3 - y4)
            except:
                # print("ValueError on L499 %g" % (self._det_(x1 - x2, y1 - y2, x3 - x4, y3 - y4)))
                y = float('Inf')
        return x, y

    def _det_(self, a, b, c, d):
        return a * d - c * b

    def interp_xy(self, x, y, atom):
        """
        given a point in u-axis, corresponding value is interpolated
        """
        f = 1.0
        interps = []
        min_distance = None
        for idx, (x1s, y1s, x2s, y2s) in enumerate(atom.sections):
            distance_1 = self._calc_distance_points_(x1s, y1s, x, y)
            distance_2 = self._calc_distance_points_(x2s, y2s, x, y)
            distance = min(distance_1, distance_2)
            if min_distance is None:
                min_distance = distance
                closest_value = atom.section_values[idx][0]
            else:
                if distance < min_distance:
                    min_distance = distance
                    value_0 = atom.section_values[idx][0]
                    value_1 = atom.section_values[idx][1]
                    closest_value = self.interpolate(x1s, y1s, x2s, y2s, x, y, value_0, value_1)
                    closest_value_0 = value_0
                    closest_value_1 = value_1
        # print closest_value,closest_value_0,closest_value_1
        return closest_value

    def interpolate(self, x1, y1, x2, y2, x3, y3, value_1, value_2):
        """
        value 1 = x1,y1
        value 2 = x2,y2
        point = x3,y3
        Interpolates linearly what is the value of point in line (x1,y1)-(x2,y2)
        """
        distance_1 = self._calc_distance_points_(x1, y1, x3, y3)
        distance_2 = self._calc_distance_points_(x2, y2, x3, y3)
        value = value_1 + (value_2 - value_1) * distance_1 / (distance_1 + distance_2)
        return value

    def parse_linestyle(self, line_style):
        """
        parses linestyle
        """
        if not re.search("solid", line_style, re.IGNORECASE) is None:
            return pyx.style.linestyle.solid
        if not re.search("dashed", line_style, re.IGNORECASE) is None:
            return pyx.style.linestyle.dashed
        if not re.search("dotted", line_style, re.IGNORECASE) is None:
            return pyx.style.linestyle.dotted
        if not re.search("dashdotted", line_style, re.IGNORECASE) is None:
            return pyx.style.linestyle.dashdotted
        # no match return default
        print("unknown linestyle: %s" % line_style)
        return pyx.style.linestyle.dashed

    def parse_linewidth(self, line_width):
        """
        parses linewidth
        """
        if not re.search("THIN", line_width) is None:
            return pyx.style.linewidth.THIN
        if not re.search("THIn", line_width) is None:
            return pyx.style.linewidth.THIn
        if not re.search("THin", line_width) is None:
            return pyx.style.linewidth.THin
        if not re.search("Thin", line_width) is None:
            return pyx.style.linewidth.Thin
        if not re.search("thin", line_width) is None:
            return pyx.style.linewidth.thin
        if not re.search("thick", line_width) is None:
            return pyx.style.linewidth.thick
        if not re.search("Thick", line_width) is None:
            return pyx.style.linewidth.Thick
        if not re.search("THick", line_width) is None:
            return pyx.style.linewidth.THick
        if not re.search("THIck", line_width) is None:
            return pyx.style.linewidth.THIck
        if not re.search("THICk", line_width) is None:
            return pyx.style.linewidth.THICk
        if not re.search("THICK", line_width) is None:
            return pyx.style.linewidth.THICK
        if not re.search("normal", line_width, re.IGNORECASE) is None:
            return pyx.style.linewidth.normal
        # no match return default
        print("unknown linewidth: %s" % line_width)
        return pyx.style.linewidth.normal

    def parse_color(self, color_str):
        """
        parses color
        """
        if re.match("GreenYellow", color_str, re.IGNORECASE):
            return pyx.color.cmyk.GreenYellow
        if re.match("Yellow", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Yellow
        if re.match("Goldenrod", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Goldenrod
        if re.match("Dandelion", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Dandelion
        if re.match("Apricot", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Apricot
        if re.match("Peach", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Peach
        if re.match("Melon", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Melon
        if re.match("YellowOrange", color_str, re.IGNORECASE):
            return pyx.color.cmyk.YellowOrange
        if re.match("Orange", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Orange
        if re.match("BurntOrange", color_str, re.IGNORECASE):
            return pyx.color.cmyk.BurntOrange
        if re.match("Bittersweet", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Bittersweet
        if re.match("RedOrange", color_str, re.IGNORECASE):
            return pyx.color.cmyk.RedOrange
        if re.match("Mahogany", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Mahogany
        if re.match("Maroon", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Maroon
        if re.match("BrickRed", color_str, re.IGNORECASE):
            return pyx.color.cmyk.BrickRed
        if re.match("Red", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Red
        if re.match("OrangeRed", color_str, re.IGNORECASE):
            return pyx.color.cmyk.OrangeRed
        if re.match("RubineRed", color_str, re.IGNORECASE):
            return pyx.color.cmyk.RubineRed
        if re.match("WildStrawberry", color_str, re.IGNORECASE):
            return pyx.color.cmyk.WildStrawberry
        if re.match("Salmon", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Salmon
        if re.match("CarnationPink", color_str, re.IGNORECASE):
            return pyx.color.cmyk.CarnationPink
        if re.match("Magenta", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Magenta
        if re.match("VioletRed", color_str, re.IGNORECASE):
            return pyx.color.cmyk.VioletRed
        if re.match("Rhodamine", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Rhodamine
        if re.match("Mulberry", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Mulberry
        if re.match("RedViolet", color_str, re.IGNORECASE):
            return pyx.color.cmyk.RedViolet
        if re.match("Fuchsia", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Fuchsia
        if re.match("Lavender", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Lavender
        if re.match("Thistle", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Thistle
        if re.match("Orchid", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Orchid
        if re.match("DarkOrchid", color_str, re.IGNORECASE):
            return pyx.color.cmyk.DarkOrchid
        if re.match("Purple", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Purple
        if re.match("Plum", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Plum
        if re.match("Violet", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Violet
        if re.match("RoyalPurple", color_str, re.IGNORECASE):
            return pyx.color.cmyk.RoyalPurple
        if re.match("BlueViolet", color_str, re.IGNORECASE):
            return pyx.color.cmyk.BlueViolet
        if re.match("Periwinkle", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Periwinkle
        if re.match("CadetBlue", color_str, re.IGNORECASE):
            return pyx.color.cmyk.CadetBlue
        if re.match("CornflowerBlue", color_str, re.IGNORECASE):
            return pyx.color.cmyk.CornflowerBlue
        if re.match("MidnightBlue", color_str, re.IGNORECASE):
            return pyx.color.cmyk.MidnightBlue
        if re.match("NavyBlue", color_str, re.IGNORECASE):
            return pyx.color.cmyk.NavyBlue
        if re.match("RoyalBlue", color_str, re.IGNORECASE):
            return pyx.color.cmyk.RoyalBlue
        if re.match("Blue", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Blue
        if re.match("Cerulean", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Cerulean
        if re.match("Cyan", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Cyan
        if re.match("ProcessBlue", color_str, re.IGNORECASE):
            return pyx.color.cmyk.ProcessBlue
        if re.match("SkyBlue", color_str, re.IGNORECASE):
            return pyx.color.cmyk.SkyBlue
        if re.match("Turquoise", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Turquoise
        if re.match("TealBlue", color_str, re.IGNORECASE):
            return pyx.color.cmyk.TealBlue
        if re.match("Aquamarine", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Aquamarine
        if re.match("BlueGreen", color_str, re.IGNORECASE):
            return pyx.color.cmyk.BlueGreen
        if re.match("Emerald", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Emerald
        if re.match("JungleGreen", color_str, re.IGNORECASE):
            return pyx.color.cmyk.JungleGreen
        if re.match("SeaGreen", color_str, re.IGNORECASE):
            return pyx.color.cmyk.SeaGreen
        if re.match("Green", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Green
        if re.match("ForestGreen", color_str, re.IGNORECASE):
            return pyx.color.cmyk.ForestGreen
        if re.match("PineGreen", color_str, re.IGNORECASE):
            return pyx.color.cmyk.PineGreen
        if re.match("LimeGreen", color_str, re.IGNORECASE):
            return pyx.color.cmyk.LimeGreen
        if re.match("YellowGreen", color_str, re.IGNORECASE):
            return pyx.color.cmyk.YellowGreen
        if re.match("SpringGreen", color_str, re.IGNORECASE):
            return pyx.color.cmyk.SpringGreen
        if re.match("OliveGreen", color_str, re.IGNORECASE):
            return pyx.color.cmyk.OliveGreen
        if re.match("RawSienna", color_str, re.IGNORECASE):
            return pyx.color.cmyk.RawSienna
        if re.match("Sepia", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Sepia
        if re.match("Brown", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Brown
        if re.match("Tan", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Tan
        if re.match("Gray", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Gray
        if re.match("Black", color_str, re.IGNORECASE):
            return pyx.color.cmyk.Black
        if re.match("White", color_str, re.IGNORECASE):
            return pyx.color.cmyk.White
        # default
        print("unknown color: %s" % color_str)
        return pyx.color.cmyk.Black

    def parse_isopleth_params(self, params):
        """
        parses single definition. Definition is for example:
          'isopleth_params':[{'color':'Black',
                              'linestyle':'Dashed',
                              'lineweight':'THick',
                              'circle_size':0.05}]
        """
        # color
        if 'color' in params:
            color_param = self.parse_color(params['color'])
        else:
            color_param = self.parse_color('black')

        # color rgb
        if 'color_rgb' in params:
            r = params['color_rgb'][0]
            g = params['color_rgb'][1]
            b = params['color_rgb'][2]
            color_param = pyx.color.rgb(r, g, b)
        # color cmyk
        if 'color_cmyk' in params:
            c = params['color_cmyk'][0]
            m = params['color_cmyk'][1]
            y = params['color_cmyk'][2]
            k = params['color_cmyk'][3]
            color_param = pyx.color.cmyk(c, m, y, k)
        # transparent
        transparent = False
        if 'transparency' in params:
            color_param_transparency = pyx.color.transparency(params['transparency'])
            transparent = True
        # linestyle
        if 'linestyle' in params:
            linestyle_param = self.parse_linestyle(params['linestyle'])
        else:
            linestyle_param = self.parse_linestyle('dashed')
        # linewidth
        if 'linewidth' in params:
            linewidth_param = self.parse_linewidth(params['linewidth'])
        else:
            linewidth_param = self.parse_linewidth('thick')
        if transparent:
            return [color_param, linestyle_param, linewidth_param, color_param_transparency]
        else:
            return [color_param, linestyle_param, linewidth_param]

    def parse_circle_size(self, params):
        #
        # circle radius
        if 'circle_size' in params:
            return params['circle_size']
        else:
            return 0.05


class Isopleth_Block_Type_1(Isopleth_Block):
    """
    type F1+F2+F3=0 isopleth
    atom stack is the stack of scales
    solution_dict is a dictionary of found solutions
    """

    def __init__(self, atom_stack, params):
        super(Isopleth_Block_Type_1, self).__init__(atom_stack, params)

    def _check_if_enough_params_(self, idx):
        """
        checks if enough numbers given to find solution
        """
        numbers = self.isopleth_values[idx]
        given = 0
        for number in numbers:
            if isinstance(number, (int, float, tuple, list)):
                given += 1
        if given < 2:
            return False  # isopleth not solvable (right now)
        else:
            return True  # isopleth solvable

    def solve(self, solutions):
        """
        solves coordinates
        solutions is list of dicts of found solutions
        """
        for idx, isopleth_values_single in enumerate(self.isopleth_values):
            if len(self.draw_coordinates) < (idx + 1):
                self.draw_coordinates.append([])  # dummy expansion of matrix
            if len(solutions) < (idx + 1):
                solutions.append({})
            if len(self.other_points) < (idx + 1):
                self.other_points.append([])
            if self._check_if_enough_params_(idx):
                x0, y0, x1, y1, x2, y2 = self.solve_single(solutions[idx],
                                                           isopleth_values_single, idx)
                self.draw_coordinates[idx] = [x0, y0, x1, y1, x2, y2]

    def solve_single(self, solution, isopleth_values, idx):
        """
        solves single isopleth
        solution = dict with values of found solutions
        isopleth_values = list of values and coordinates
        idx = # of isopleth line
        """
        atom_stack = self.atom_stack
        f1_known = False
        f2_known = False
        f3_known = False
        # f1 known
        if isinstance(isopleth_values[0], (int, float)):
            x0 = atom_stack[0].give_x(isopleth_values[0])
            y0 = atom_stack[0].give_y(isopleth_values[0])
            f1_known = True
        if isinstance(isopleth_values[0], tuple):
            x0 = isopleth_values[0][0]
            y0 = isopleth_values[0][1]
            f1_known = True
        if isinstance(isopleth_values[0], list):  # = this is grid
            x0 = atom_stack[0].give_x_grid(isopleth_values[0][0], isopleth_values[0][1])
            y0 = atom_stack[0].give_y_grid(isopleth_values[0][0], isopleth_values[0][1])
            f1_known = True
        # f2 known
        if isinstance(isopleth_values[1], (int, float)):
            x1 = atom_stack[1].give_x(isopleth_values[1])
            y1 = atom_stack[1].give_y(isopleth_values[1])
            f2_known = True
        if isinstance(isopleth_values[1], tuple):
            x1 = isopleth_values[1][0]
            y1 = isopleth_values[1][1]
            f2_known = True
        if isinstance(isopleth_values[1], list):  # = this is grid
            x1 = atom_stack[1].give_x_grid(isopleth_values[1][0], isopleth_values[1][1])
            y1 = atom_stack[1].give_y_grid(isopleth_values[1][0], isopleth_values[1][1])
            f2_known = True
        # f3 known
        if isinstance(isopleth_values[2], (int, float)):
            x2 = atom_stack[2].give_x(isopleth_values[2])
            y2 = atom_stack[2].give_y(isopleth_values[2])
            f3_known = True
        if isinstance(isopleth_values[2], tuple):
            x2 = isopleth_values[2][0]
            y2 = isopleth_values[2][1]
            f3_known = True
        if isinstance(isopleth_values[2], list):  # = this is grid
            x2 = atom_stack[2].give_x_grid(isopleth_values[2][0], isopleth_values[2][1])
            y2 = atom_stack[2].give_y_grid(isopleth_values[2][0], isopleth_values[2][1])
            f3_known = True
        if not f1_known:
            # line=self.atom_stack[0].line
            x0, y0 = self._find_closest_point_(self.atom_stack[0].sections, x1, y1, x2, y2)
            other_points = self._find_closest_other_points_(self.atom_stack[0].sections, x1, y1, x2, y2, x0, y0)
            # solution[isopleth_values[0]]=(x0,y0)
            if not self.atom_stack[0].params['tag'] == 'none':
                solution[self.atom_stack[0].params['tag']] = (x0, y0)
            isopleth_values[0] = (x0, y0)
            self.other_points[idx].append(other_points)
        if not f2_known:
            # line=self.atom_stack[1].line
            x1, y1 = self._find_closest_point_(self.atom_stack[1].sections, x0, y0, x2, y2)
            other_points = self._find_closest_other_points_(self.atom_stack[1].sections, x0, y0, x2, y2, x1, y1)
            # solution[isopleth_values[1]]=(x1,y1)
            if not self.atom_stack[1].params['tag'] == 'none':
                solution[self.atom_stack[1].params['tag']] = (x1, y1)
            isopleth_values[1] = (x1, y1)
            self.other_points[idx].append(other_points)
        if not f3_known:
            # line=self.atom_stack[2].line
            x2, y2 = self._find_closest_point_(self.atom_stack[2].sections, x0, y0, x1, y1)
            other_points = self._find_closest_other_points_(self.atom_stack[2].sections, x0, y0, x1, y1, x2, y2)
            # solution[isopleth_values[2]]=(x2,y2)
            if not self.atom_stack[2].params['tag'] == 'none':
                solution[self.atom_stack[2].params['tag']] = (x2, y2)
            isopleth_values[2] = (x2, y2)
            self.other_points[idx].append(other_points)
        return x0, y0, x1, y1, x2, y2


class Isopleth_Block_Type_2(Isopleth_Block_Type_1):
    """
    type F1=F2*F3 isopleth
    atom stack is the stack of scales
    solution_dict is a dictionary of found solutions
    """

    def __init__(self, atom_stack, params):
        super(Isopleth_Block_Type_2, self).__init__(atom_stack, params)


class Isopleth_Block_Type_5(Isopleth_Block):
    """
    type 5 = contour, nomo_block = block of class Nomo_Block_Type_5
    """

    def __init__(self, atom_stack, params, nomo_block):
        super(Isopleth_Block_Type_5, self).__init__(atom_stack, params)
        self.nomo_block = nomo_block

    def _check_if_enough_params_(self, idx):
        """
        checks if enough numbers given to find solution
        """
        numbers = self.isopleth_values[idx]
        given = 0
        for number in numbers:
            if isinstance(number, (int, float, tuple)):
                given = given + 1
        if given < 2:
            return False  # isopleth not solvable (right now)
        else:
            return True  # isopleth solvable

    def draw(self, canvas, draw_params=[{}]):
        """
        draws the isopleth
        """
        for idx, (x1, y1, x2, y2, x3, y3) in enumerate(self.draw_coordinates):
            if len(draw_params) > idx:
                p = draw_params[idx]
            else:
                p = draw_params[len(draw_params) - 1]
            draw_params_list = self.parse_isopleth_params(p)
            color_param = draw_params_list[0]
            circle_radius = self.parse_circle_size(p)
            canvas.stroke(pyx.path.line(x1, y1, x2, y2), draw_params_list)
            canvas.stroke(pyx.path.line(x2, y2, x3, y3), draw_params_list)
            self._draw_circle_(canvas, x1, y1, circle_radius, color_param)
            self._draw_circle_(canvas, x2, y2, circle_radius, color_param)
            self._draw_circle_(canvas, x3, y3, circle_radius, color_param)
        for idx, line_points in enumerate(self.other_points):
            if len(draw_params) > idx:
                p = draw_params[idx]
            else:
                p = draw_params[len(draw_params) - 1]
            draw_params_list = self.parse_isopleth_params(p)
            color_param = draw_params_list[0]
            circle_radius = self.parse_circle_size(p)
            for points in line_points:
                for (x, y) in points:
                    self._draw_circle_(canvas, x, y, circle_radius)

    def solve(self, solutions):
        """
        solves coordinates
        solutions is list of dicts of found solutions
        """
        for idx, isopleth_values_single in enumerate(self.isopleth_values):
            if len(self.draw_coordinates) < (idx + 1):
                self.draw_coordinates.append([])  # dummy expansion of matrix
            if len(solutions) < (idx + 1):
                solutions.append({})
            if len(self.other_points) < (idx + 1):
                self.other_points.append([])
            if self._check_if_enough_params_(idx):
                x_u, y_u, x_v, y_v, x_wd, y_wd = self.solve_single(solutions[idx],
                                                                   isopleth_values_single, idx)
                self.draw_coordinates[idx] = [x_u, y_u, x_v, y_v, x_wd, y_wd]

    def solve_single(self, solution, isopleth_values, idx):
        """
        solves the thing
        """
        atom_stack = self.atom_stack
        u_known = False
        v_known = False
        wd_known = False
        # initial values that are replaced (for debugging)
        x_u = -10
        y_u = -10
        x_wd = -10
        y_wd = -10
        x_v = -10
        x_v = -10
        # u known
        if isinstance(isopleth_values[0], (int, float)):
            x_u, x_u_ini, y_u_ini = self.x_u(isopleth_values[0])
            y_u, x_u_ini, y_u_ini = self.y_u(isopleth_values[0])
            u_known = True
        # u known as tuple
        if isinstance(isopleth_values[0], tuple):
            u_value = self.u_x_y_interp(isopleth_values[0][0], isopleth_values[0][1])
            x_u, x_u_ini, y_u_ini = self.x_u(u_value)
            y_u, x_u_ini, y_u_ini = self.y_u(u_value)
            u_known = True
        # wd known
        if isinstance(isopleth_values[2], (int, float)):
            x_wd, x_wd_ini, y_wd_ini = self.x_wd(isopleth_values[2])
            y_wd, x_wd_ini, y_wd_ini = self.y_wd(isopleth_values[2])
            wd_known = True
        # wd known as tuple
        if isinstance(isopleth_values[2], tuple):
            wd_value = self.wd_x_y_interp(isopleth_values[2][0], isopleth_values[2][1])
            x_wd, x_wd_ini, y_wd_ini = self.x_wd(wd_value)
            y_wd, x_wd_ini, y_wd_ini = self.y_wd(wd_value)
            wd_known = True
        # v known as tuple (possible?)
        if isinstance(isopleth_values[1], tuple):
            x_v, x_v_ini, y_v_ini = isopleth_values[1][0]
            y_v, x_v_ini, y_v_ini = isopleth_values[1][1]
            v_known = True
        # v and wd known as values
        if isinstance(isopleth_values[1], (int, float)) and isinstance(isopleth_values[2], (int, float)):
            x_v, x_v_ini, y_v_ini = self.x_v_wd(isopleth_values[1], isopleth_values[2])
            y_v, x_v_ini, y_v_ini = self.y_v_wd(isopleth_values[1], isopleth_values[2])
            v_known = True
        # v known as value and wd known as tuple (x,y)
        if isinstance(isopleth_values[1], (int, float)) and isinstance(isopleth_values[2], (tuple)):
            x_v, x_v_ini, y_v_ini = self.x_v_wd_tuple(isopleth_values[1], isopleth_values[2])
            y_v, x_v_ini, y_v_ini = self.y_v_wd_tuple(isopleth_values[1], isopleth_values[2])
            v_known = True
        # v and u known as values
        if isinstance(isopleth_values[1], (int, float)) and isinstance(isopleth_values[0], (int, float)):
            x_v, y_v, x_v_ini, y_v_ini = self.xy_v_u(isopleth_values[1], isopleth_values[0])
            v_known = True
        # v known as value and u known as tuple (x,y)
        if isinstance(isopleth_values[1], (int, float)) and isinstance(isopleth_values[0], (tuple)):
            x_v, y_v, x_v_ini, y_v_ini = self.xy_v_u_tuple(isopleth_values[1], isopleth_values[0])
            v_known = True
        # now, all needed coordinates known
        if not u_known and v_known and wd_known:
            x_u_ini = self.nomo_block.grid_box.params_u['F'](0)
            y_u_ini = y_v_ini
            x_u = self.nomo_block._give_trafo_x_(x_u_ini, y_u_ini)
            y_u = self.nomo_block._give_trafo_y_(x_u_ini, y_u_ini)
            if not self.nomo_block.grid_box.params_u['tag'] == 'none':
                solution[self.nomo_block.grid_box.params_u['tag']] = (x_u, y_u)
        if not v_known and u_known and wd_known:
            x_v_ini = x_wd_ini
            y_v_ini = y_u_ini
            x_v = self.nomo_block._give_trafo_x_(x_v_ini, y_v_ini)
            y_v = self.nomo_block._give_trafo_y_(x_v_ini, y_v_ini)
            if not self.nomo_block.grid_box.params_v['tag'] == 'none':
                solution[self.nomo_block.grid_box.params_v['tag']] = (x_v, y_v)
        if not wd_known and u_known and v_known:
            x_wd_ini = x_v_ini
            y_wd_ini = self.nomo_block.grid_box.params_wd['G'](0)
            x_wd = self.nomo_block._give_trafo_x_(x_wd_ini, y_wd_ini)
            y_wd = self.nomo_block._give_trafo_y_(x_wd_ini, y_wd_ini)
            if not self.nomo_block.grid_box.params_wd['tag'] == 'none':
                solution[self.nomo_block.grid_box.params_wd['tag']] = (x_wd, y_wd)
        return x_u, y_u, x_v, y_v, x_wd, y_wd

    def x_u(self, u):
        """
        give x(u)
        """
        x0 = self.nomo_block.grid_box.params_u['F'](u)
        y0 = self.nomo_block.grid_box.params_u['G'](u)
        return self.nomo_block._give_trafo_x_(x0, y0), x0, y0

    def y_u(self, u):
        """
        give y(u)
        """
        x0 = self.nomo_block.grid_box.params_u['F'](u)
        y0 = self.nomo_block.grid_box.params_u['G'](u)
        return self.nomo_block._give_trafo_y_(x0, y0), x0, y0

    def x_wd(self, wd):
        """
        give x(wd)
        """
        x0 = self.nomo_block.grid_box.params_wd['F'](wd)
        y0 = self.nomo_block.grid_box.params_wd['G'](wd)
        return self.nomo_block._give_trafo_x_(x0, y0), x0, y0

    def y_wd(self, wd):
        """
        give y(wd)
        """
        x0 = self.nomo_block.grid_box.params_wd['F'](wd)
        y0 = self.nomo_block.grid_box.params_wd['G'](wd)
        return self.nomo_block._give_trafo_y_(x0, y0), x0, y0

    def xy_v_u(self, v, u):
        """
        give x(v,u), y(v,u)
        """
        x_start = self.nomo_block.grid_box.x_left
        x_stop = self.nomo_block.grid_box.x_right
        if x_start > x_stop:  # should no be
            x_start, x_stop = x_stop, x_start
        # x_init=(x_start+x_stop)/2.0
        v_func = self.nomo_block.grid_box.v_func
        u_func = self.nomo_block.grid_box.u_func
        u_value = u_func(u)  # = y
        func_opt = lambda x: (v_func(x, v) - u_value) ** 2  # func to minimize
        # let's try to find good starting point for optimization
        x_range = arange(x_start, x_stop, (x_stop - x_start) / 30.0, dtype=complex)
        #        print "x_range:"
        #        print x_range
        # use complex numbers to filter results with complex part
        # values = func_opt(x_range.astype(complex))
        values = x_range
        values_list_complex = values.tolist()
        values_list = []
        for value in values_list_complex:
            if value.imag == 0:
                values_list.append(value.real)
            else:
                values_list.append(math.inf)  # large number
            #        print "values_list:"
            #        print values_list
        min_x_idx = values_list.index(min(values_list))
        x_init = x_range[min_x_idx].real
        #        print "x_start %g"%x_start
        #        print "x_stop %g"%x_stop
        #        print "x_init %g"%x_init
        # find x point where u meets v = optimization
        x_opt = scipy.optimize.fmin(func_opt, [x_init], disp=0, maxiter=1e5, maxfun=1e5, ftol=1e-8, xtol=1e-8)[0]
        x_transformed = self.nomo_block._give_trafo_x_(x_opt, u_value)
        y_transformed = self.nomo_block._give_trafo_y_(x_opt, u_value)
        return x_transformed, y_transformed, x_opt, u_value

    def xy_v_u_tuple(self, v, u):
        """
        give x(v,u), y(v,u) where u is tuple (x,y)
        """
        return self.xy_v_u(v, self.u_x_y_interp(u[0], u[1]))

    def x_v_wd(self, v, wd):
        """
        give x(v,wd)
        """
        x0 = self.nomo_block.grid_box.params_wd['F'](wd)
        y0 = self.nomo_block.grid_box.v_func(x0, v)
        return self.nomo_block._give_trafo_x_(x0, y0), x0, y0

    def y_v_wd(self, v, wd):
        """
        give y(v,wd)
        """
        x0 = self.nomo_block.grid_box.params_wd['F'](wd)
        y0 = self.nomo_block.grid_box.v_func(x0, v)
        return self.nomo_block._give_trafo_y_(x0, y0), x0, y0

    def x_v_wd_tuple(self, v, wd):
        """
        give x(v,wd) where wd is tuple of final coordinates
        """
        return self.x_v_wd(v, self.wd_x_y_interp(wd[0], wd[1]))

    def y_v_wd_tuple(self, v, wd):
        """
        give y(v,wd) where wd is tuple of final coordinates
        """
        return self.y_v_wd(v, self.wd_x_y_interp(wd[0], wd[1]))

    def wd_x_y_interp(self, x, y):
        """
        given a point in wd-axis, corresponding value is interpolated
        """
        f = 1.0
        interps = []
        min_distance = None
        for idx, (x1s, y1s, x2s, y2s) in enumerate(self.nomo_block.atom_wd.sections):
            distance_1 = self._calc_distance_points_(x1s, y1s, x, y)
            distance_2 = self._calc_distance_points_(x2s, y2s, x, y)
            distance = min(distance_1, distance_2)
            if min_distance is None:
                min_distance = distance
                closest_value = self.nomo_block.atom_wd.section_values[idx][0]
            else:
                if distance < min_distance:
                    min_distance = distance
                    value_0 = self.nomo_block.atom_wd.section_values[idx][0]
                    value_1 = self.nomo_block.atom_wd.section_values[idx][1]
                    closest_value = self.interpolate(x1s, y1s, x2s, y2s, x, y, value_0, value_1)
                    closest_value_0 = value_0
                    closest_value_1 = value_1
        # print closest_value,closest_value_0,closest_value_1
        return closest_value

    def u_x_y_interp(self, x, y):
        """
        given a point in u-axis, corresponding value is interpolated
        """
        f = 1.0
        interps = []
        min_distance = None
        for idx, (x1s, y1s, x2s, y2s) in enumerate(self.nomo_block.atom_u.sections):
            distance_1 = self._calc_distance_points_(x1s, y1s, x, y)
            distance_2 = self._calc_distance_points_(x2s, y2s, x, y)
            distance = min(distance_1, distance_2)
            if min_distance is None:
                min_distance = distance
                closest_value = self.nomo_block.atom_u.section_values[idx][0]
            else:
                if distance < min_distance:
                    min_distance = distance
                    value_0 = self.nomo_block.atom_u.section_values[idx][0]
                    value_1 = self.nomo_block.atom_u.section_values[idx][1]
                    closest_value = self.interpolate(x1s, y1s, x2s, y2s, x, y, value_0, value_1)
                    closest_value_0 = value_0
                    closest_value_1 = value_1
        # print closest_value,closest_value_0,closest_value_1
        return closest_value


class Isopleth_Block_Type_6(Isopleth_Block):
    """
    type single
    atom stack is the stack of scales
    solution_dict is a dictionary of found solutions
    """

    def __init__(self, atom_stack, params):
        super(Isopleth_Block_Type_6, self).__init__(atom_stack, params)

    def _check_if_enough_params_(self, idx):
        """
        checks if enough numbers given to find solution
        """
        numbers = self.isopleth_values[idx]
        given = 0
        for number in numbers:
            if isinstance(number, (int, float, tuple, list)):
                given = given + 1
        if given > 0:
            return True  # isopleth solvable (right now)
        else:
            return False  # isopleth not solvable

    def solve(self, solutions):
        """
        solves coordinates
        solutions is list of dicts of found solutions
        """
        for idx, isopleth_values_single in enumerate(self.isopleth_values):
            if len(self.draw_coordinates) < (idx + 1):
                self.draw_coordinates.append([])  # dummy expansion of matrix
            if len(solutions) < (idx + 1):
                solutions.append({})
            if len(self.other_points) < (idx + 1):
                self.other_points.append([])
            if self._check_if_enough_params_(idx):
                x1, y1, x2, y2 = self.solve_single(solutions[idx], isopleth_values_single, idx)
                self.draw_coordinates[idx] = [x1, y1, x2, y2]

    def solve_single(self, solution, isopleth_values, idx):
        """
        solves single isopleth
        solution = dict with values of found solutions
        isopleth_values = list of values and coordinates
        idx = # of isopleth line
        """
        atom_stack = self.atom_stack
        atom1 = atom_stack[0]
        atom2 = atom_stack[1]
        f1_known = False
        f2_known = False
        # f1 value known
        if isinstance(isopleth_values[0], (int, float)):
            x1 = atom_stack[0].give_x(isopleth_values[0])
            y1 = atom_stack[0].give_y(isopleth_values[0])
            f1_known = True
            value = self.interp_xy(x1, y1, atom1)
        if isinstance(isopleth_values[0], tuple):
            x1 = isopleth_values[0][0]
            y1 = isopleth_values[0][1]
            f1_known = True
            value = self.interp_xy(x1, y1, atom1)
        # f2 value known
        if isinstance(isopleth_values[1], (int, float)):
            x2 = atom_stack[1].give_x(isopleth_values[1])
            y2 = atom_stack[1].give_y(isopleth_values[1])
            f2_known = True
            value = self.interp_xy(x2, y2, atom2)
        if isinstance(isopleth_values[1], tuple):
            x2 = isopleth_values[1][0]
            y2 = isopleth_values[1][1]
            f2_known = True
            value = self.interp_xy(x2, y2, atom2)
        if not f1_known:
            x1 = atom1.give_x(value)
            y1 = atom1.give_y(value)
        if not f2_known:
            x2 = atom2.give_x(value)
            y2 = atom2.give_y(value)
        # let's save the result
        if not atom1.params['tag'] == 'none':
            solution[atom1.params['tag']] = (x1, y1)
        if not atom2.params['tag'] == 'none':
            solution[atom2.params['tag']] = (x2, y2)
        return x1, y1, x2, y2

    def draw(self, canvas, draw_params=[{}]):
        """
        draws the isopleth
        """
        for idx, (x1, y1, x2, y2) in enumerate(self.draw_coordinates):
            if len(draw_params) > idx:
                p = draw_params[idx]
            else:
                p = draw_params[len(draw_params) - 1]
            draw_params_list = self.parse_isopleth_params(p)
            color_param = draw_params_list[0]
            circle_radius = self.parse_circle_size(p)
            x_offset1 = self.atom_stack[0].params['align_x_offset']
            y_offset1 = self.atom_stack[0].params['align_y_offset']
            x_offset2 = self.atom_stack[1].params['align_x_offset']
            y_offset2 = self.atom_stack[1].params['align_y_offset']
            canvas.stroke(pyx.path.line(x1 - x_offset1, y1 - y_offset1, x2 - x_offset2, y2 - y_offset2),
                          draw_params_list)
            self._draw_circle_(canvas, x1, y1, circle_radius, color_param)
            self._draw_circle_(canvas, x2, y2, circle_radius, color_param)
        for idx, line_points in enumerate(self.other_points):
            if len(draw_params) > idx:
                p = draw_params[idx]
            else:
                p = draw_params[len(draw_params) - 1]
            draw_params_list = self.parse_isopleth_params(p)
            color_param = draw_params_list[0]
            circle_radius = self.parse_circle_size(p)
            for points in line_points:
                for (x, y) in points:
                    self._draw_circle_(canvas, x, y, circle_radius, color_param)


class Isopleth_Block_Type_7(Isopleth_Block_Type_1):
    """
    type 1/F1+1/F2=1/F3 isopleth
    atom stack is the stack of scales
    solution_dict is a dictionary of found solutions
    """

    def __init__(self, atom_stack, params):
        super(Isopleth_Block_Type_7, self).__init__(atom_stack, params)


class Isopleth_Block_Type_8(Isopleth_Block):
    """
    type single
    atom stack is the stack of scales
    solution_dict is a dictionary of found solutions
    """

    def __init__(self, atom_stack, params):
        super(Isopleth_Block_Type_8, self).__init__(atom_stack, params)

    def _check_if_enough_params_(self, idx):
        """
        checks if enough numbers given to find solution
        """
        numbers = self.isopleth_values[idx]
        given = 0
        for number in numbers:
            if isinstance(number, (int, float, tuple, list)):
                given = given + 1
        if given > 0:
            return True  # isopleth solvable (right now)
        else:
            return False  # isopleth not solvable

    def solve(self, solutions):
        """
        solves coordinates
        solutions is list of dicts of found solutions
        """
        for idx, isopleth_values_single in enumerate(self.isopleth_values):
            if len(self.draw_coordinates) < (idx + 1):
                self.draw_coordinates.append([])  # dummy expansion of matrix
            if len(solutions) < (idx + 1):
                solutions.append({})
            if len(self.other_points) < (idx + 1):
                self.other_points.append([])
            if self._check_if_enough_params_(idx):
                x0, y0 = self.solve_single(solutions[idx], isopleth_values_single, idx)
                self.draw_coordinates[idx] = [x0, y0]

    def solve_single(self, solution, isopleth_values, idx):
        """
        solves single isopleth
        solution = dict with values of found solutions
        isopleth_values = list of values and coordinates
        idx = # of isopleth line
        """
        atom_stack = self.atom_stack
        # value known
        if isinstance(isopleth_values[0], (int, float)):
            x0 = atom_stack[0].give_x(isopleth_values[0])
            y0 = atom_stack[0].give_y(isopleth_values[0])
        if isinstance(isopleth_values[0], tuple):
            x0 = isopleth_values[0][0]
            y0 = isopleth_values[0][1]
        if not self.atom_stack[0].params['tag'] == 'none':
            solution[self.atom_stack[0].params['tag']] = (x0, y0)
        return x0, y0

    def draw(self, canvas, draw_params=[{}]):
        """
        draws the isopleth
        """
        for idx, (x1, y1) in enumerate(self.draw_coordinates):
            if len(draw_params) > idx:
                p = draw_params[idx]
            else:
                p = draw_params[len(draw_params) - 1]
            draw_params_list = self.parse_isopleth_params(p)
            color_param = draw_params_list[0]
            circle_radius = self.parse_circle_size(p)
            x_offset = self.atom_stack[0].params['align_x_offset']
            y_offset = self.atom_stack[0].params['align_y_offset']
            if x_offset != 0 or y_offset != 0:
                canvas.stroke(pyx.path.line(x1, y1, x1 - x_offset, y1 - y_offset), draw_params_list)
            self._draw_circle_(canvas, x1, y1, circle_radius, color_param)
        for idx, line_points in enumerate(self.other_points):
            if len(draw_params) > idx:
                p = draw_params[idx]
            else:
                p = draw_params[len(draw_params) - 1]
            draw_params_list = self.parse_isopleth_params(p)
            color_param = draw_params_list[0]
            circle_radius = self.parse_circle_size(p)
            for points in line_points:
                for (x, y) in points:
                    self._draw_circle_(canvas, x, y, circle_radius, color_param)


class Isopleth_Block_Type_9(Isopleth_Block_Type_1):
    """
    type general determinant isopleth
    note, that parameters for grid should be given, they are not solved
    """

    def __init__(self, atom_stack, params):
        super(Isopleth_Block_Type_9, self).__init__(atom_stack, params)


class Isopleth_Block_Type_10(Isopleth_Block_Type_1):
    """
    type F1(u)+F2(v)*F3(w)+F4(w)=0
    atom stack is the stack of scales
    solution_dict is a dictionary of found solutions
    """

    def __init__(self, atom_stack, params):
        super(Isopleth_Block_Type_10, self).__init__(atom_stack, params)



# -*- coding: utf-8 -*-
#
#    This file is part of PyNomo -
#    a program to create nomographs with Python (https://github.com/lefakkomies/pynomo)
#
#    Copyright (C) 2007-2019  Leif Roschier  <lefakkomies@users.sourceforge.net>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import math
import numpy as np
from scipy import linalg
#import scipy


class FourPoint(object):
    """
    class to transform four points to four points.
    """

    def __init__(self, x1, y1, x2, y2, x3, y3, x4, y4, x1d, y1d, x2d, y2d, x3d, y3d, x4d, y4d):
        """
        assumes:
           (x1,y1)     (x3,y3)          (x1d,y1d)      (x3d,y3d)
               |  polygon  |      ---->      |   polygon  |
            (x2,y2)     (x4,y4)          (x2d,y2d)      (x4d,y4d)
        """
        trafo_1 = self.find_trafo_to_unity_rectangle(x1, y1, x2, y2, x3, y3, x4, y4)
        trafo_2 = self.find_trafo_to_unity_rectangle(x1d, y1d, x2d, y2d, x3d, y3d, x4d, y4d)
        #        test1=np.array([[x1,y1,1.0],
        #                     [x2,y2,1.0],
        #                     [x3,y3,1.0]])
        #        test2=np.array([[x1,y1,1.0],
        #                     [x2,y2,1.0],
        #                     [x4,y4,1.0]])
        #        test1d=np.array([[x1d,y1d,1.0],
        #                     [x2d,y2d,1.0],
        #                     [x3d,y3d,1.0]])
        #        test2d=np.array([[x1d,y1d,1.0],
        #                     [x2d,y2d,1.0],
        #                     [x4d,y4d,1.0]])
        #        test1f=np.dot(test1,trafo_1)
        #        test2f=np.dot(test2,trafo_1)
        #        test1fd=np.dot(test1d,trafo_2)
        #        test2fd=np.dot(test2d,trafo_2)
        #        print test1f
        #        print test2f
        #        print test1fd
        #        print test2fd
        #        print "goal"
        #        print x1d,y1d,x2d,y2d,x3d,y3d,x4d,y4d
        self.trafo = np.dot(trafo_1, linalg.inv(trafo_2))
        #        print "result"
        #        print np.dot(test1,self.trafo)
        #        print np.dot(test2,self.trafo)
        """
        o1,o2,o3,p1,p2,p3,q1,q2,q3=self.find_three_points_to_transform(x1,y1,x2,y2,x3,y3,x4,y4)
        self.transformation=self.find_transformation_points_to_rectangle(o1,o2,o3,p1,p2,p3,q1,q2,q3)
        initial=np.array([[x1,y1,1.0],
                       [x2,y2,1.0],
                       [x3,y3,1.0]])
        initial2=np.array([[x1,y1,1.0],
                       [x2,y2,1.0],
                       [x4,y4,1.0]])
        inter=np.dot(initial,self.transformation)
        xi1=inter[0][0]/inter[0][2]
        yi1=inter[0][1]/inter[0][2]
        xi2=inter[1][0]/inter[1][2]
        yi2=inter[1][1]/inter[1][2]
        xi3=inter[2][0]/inter[2][2]
        yi3=inter[2][1]/inter[2][2]
        print xi1,yi1,xi2,yi2,xi3,yi3

        self.transformation1=self.affine_trafo_3_points(xi1,yi1,xi2,yi2,xi3,yi3,
                                                        0.0,0.0,0.0,1.0,1.0,0.0)
        self.total_trafo=np.dot(self.transformation,self.transformation1)
        final2=np.dot(initial2,self.transformation)
        final1a=np.dot(initial,self.total_trafo)
        print np.dot(np.array([2,2,1]),self.transformation)
        print np.dot(np.array([2,4,1]),self.transformation)
        print "initial"
        print initial
        print "inter"
        print inter
        print "final2"
        print final2
        print "transformation"
        print self.transformation
        print "final1a"
        print final1a
        """

    def give_trafo_mat(self):
        """
        gives transformation matrix
        """
        mat = self.trafo
        return mat[0][0], mat[1][0], mat[2][0], \
               mat[0][1], mat[1][1], mat[2][1], \
               mat[0][2], mat[1][2], mat[2][2]

    def find_trafo_to_unity_rectangle(self, x1, y1, x2, y2, x3, y3, x4, y4):
        """
        transforms 4 points to unity rectangle
        point 1 -> (0,0)
        point 2 -> (0,1)
        point 3 -> (1,0)
        point 4 -> (1,1)
        """
        # permute if quadilateral not in correct order
        x1p, y1p, x2p, y2p, x3p, y3p, x4p, y4p = self.check_order(x1, y1, x2, y2, x3, y3, x4, y4)
        # find 3 points out of 4 points to transform
        o1, o2, o3, p1, p2, p3, q1, q2, q3 = self.find_three_points_to_transform(x1p, y1p, x2p, y2p, x3p, y3p, x4p, y4p)
        transformation_1 = self.find_transformation_points_to_rectangle(o1, o2, o3, p1, p2, p3, q1, q2, q3)
        # three points to make correct order with affine transformation
        initial = np.array([[x1, y1, 1.0],
                         [x2, y2, 1.0],
                         [x3, y3, 1.0]])
        # intermediate points
        inter = np.dot(initial, transformation_1)
        xi1 = inter[0][0] / inter[0][2]
        yi1 = inter[0][1] / inter[0][2]
        xi2 = inter[1][0] / inter[1][2]
        yi2 = inter[1][1] / inter[1][2]
        xi3 = inter[2][0] / inter[2][2]
        yi3 = inter[2][1] / inter[2][2]
        transformation_2 = self.affine_trafo_3_points(xi1, yi1, xi2, yi2, xi3, yi3,
                                                      0.0, 0.0, 0.0, 1.0, 1.0, 0.0)
        total_trafo = np.dot(transformation_1, transformation_2)
        return total_trafo

    def calc_distance_points(self, x1, y1, x2, y2):
        """
        calcs distance between two points
        """
        return np.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2)

    def check_order(self, x1, y1, x2, y2, x3, y3, x4, y4):
        """
        makes sure that quadrilateral has order 1-2-4-3
        calculates angles around
        """

        def angle(v1, v2):
            v1_l = np.sqrt(v1[0] ** 2 + v1[1] ** 2)
            v2_l = np.sqrt(v2[0] ** 2 + v2[1] ** 2)
            return math.acos(np.dot(v1, v2) / (v1_l * v2_l))

        v21 = np.array([x2 - x1, y2 - y1])
        v42 = np.array([x4 - x2, y4 - y2])
        v34 = np.array([x3 - x4, y3 - y4])
        v13 = np.array([x1 - x3, y1 - y3])
        angle_1 = angle(v21, -v42)
        angle_2 = angle(v42, -v34)
        angle_3 = angle(v34, -v13)
        angle_4 = angle(v13, -v21)
        #print (angle_1 + angle_2 + angle_3 + angle_4)
        if abs((angle_1 + angle_2 + angle_3 + angle_4) - 2 * math.pi) < 1e-9:
            return x1, y1, x2, y2, x3, y3, x4, y4
        else:
            return x1, y1, x2, y2, x4, y4, x3, y3

    def check_if_both_lines_collinear(self, x1, y1, x2, y2, x3, y3, x4, y4):
        """
        checks if both opposite edeges collinear
        """
        if self.collinear(x1, y1, x2, y2, x3, y3, x4, y4) and \
                self.collinear(x1, y1, x3, y3, x2, y2, x4, y4):
            return True
        else:
            return False

    def collinear(self, x1, y1, x2, y2, x3, y3, x4, y4):
        """
        checks if lines (x1,y1)-(x2,y2) and (x3,y3)-(x4,y4) collinear
        """

        def det(a, b, c, d):
            return a * d - c * b

        determinant = det(x1 - x2, y1 - y2, x3 - x4, y3 - y4)
        # print "determinant"
        # print determinant
        if abs(determinant) < 1e-9:
            return True
        else:
            return False

    def two_line_intersection(self, x1, y1, x2, y2, x3, y3, x4, y4):
        """
        intersection of lines (x1,y1)-(x2,y2) and (x3,y3)-(x4,y4)
        """

        def det(a, b, c, d):
            return a * d - c * b

        x = det(det(x1, y1, x2, y2), (x1 - x2), det(x3, y3, x4, y4), (x3 - x4)) / \
            det(x1 - x2, y1 - y2, x3 - x4, y3 - y4)
        y = det(det(x1, y1, x2, y2), (y1 - y2), det(x3, y3, x4, y4), (y3 - y4)) / \
            det(x1 - x2, y1 - y2, x3 - x4, y3 - y4)
        return x, y

    def infinity_point(self, x1, y1, x2, y2):
        """
        finds point of line (of two parallel lines) end at infinity
        """
        x = x2 - x1
        y = y2 - y1
        d = 0.0  # inifinity
        return x, y, d

    def find_three_points_to_transform(self, x1, y1, x2, y2, x3, y3, x4, y4):
        """
        ...of points(x1,y1),(x2,y2),(x3,y3),(x4,y4)
        Three points are (o1,o2,o3),(p1,p2,p3),(q1,q2,q3)
        """
        # o
        # o1,o2,o3=x1,y1,1.0
        # P
        # check if lines (x1,y1)-(x2,y2) and (x3,y3)-(x4,y4) collinear
        if self.collinear(x1, y1, x2, y2, x3, y3, x4, y4):
            p1, p2, p3 = self.infinity_point(x1, y1, x2, y2)
            potential_points_1 = [1, 3]
        else:
            # lines 1-2, 3-4
            p1, p2 = self.two_line_intersection(x1, y1, x2, y2, x3, y3, x4, y4)
            p3 = 1.0  # divider of x1,y1
            if self.calc_distance_points(p1, p2, x1, y1) < self.calc_distance_points(p1, p2, x2, y2):
                potential_points_1 = [2, 4]
            else:
                potential_points_1 = [1, 3]
        # Q
        # check if lines (x1,y1)-(x3,y3) and (x2,y2)-(x4,y4) collinear
        if self.collinear(x1, y1, x3, y3, x2, y2, x4, y4):
            q1, q2, q3 = self.infinity_point(x1, y1, x3, y3)
            potential_points_2 = [1, 2]
        else:
            # lines 1-3, 2-4
            q1, q2 = self.two_line_intersection(x1, y1, x3, y3, x2, y2, x4, y4)
            q3 = 1.0  # divider of x1,y1
            if self.calc_distance_points(q1, q2, x1, y1) < self.calc_distance_points(q1, q2, x3, y3):
                potential_points_2 = [3, 4]
            else:
                potential_points_2 = [1, 2]

        if potential_points_1.count(1) == 1 and potential_points_2.count(1) == 1:
            o1, o2, o3 = x1, y1, 1.0
        if potential_points_1.count(2) == 1 and potential_points_2.count(2) == 1:
            o1, o2, o3 = x2, y2, 1.0
        if potential_points_1.count(3) == 1 and potential_points_2.count(3) == 1:
            o1, o2, o3 = x3, y3, 1.0
        if potential_points_1.count(4) == 1 and potential_points_2.count(4) == 1:
            o1, o2, o3 = x4, y4, 1.0
        return o1, o2, o3, p1, p2, p3, q1, q2, q3

    def find_transformation_points_to_rectangle(self, o1, o2, o3, p1, p2, p3, q1, q2, q3):
        """
        finds transformation that transforms points o,p,q to
        orthogonal rectangle. See Otto: p.40
        """
        trafo_mat_inv = np.array([[p1, p2, p3],
                               [q1, q2, q3],
                               [o1, o2, o3]])
        # print "trafo_mat_inv"
        # print trafo_mat_inv
        trafo_mat = linalg.inv(trafo_mat_inv)
        return trafo_mat

    def affine_trafo_3_points(self, x1, y1, x2, y2, x3, y3, x1d, y1d, x2d, y2d, x3d, y3d):
        """
        transforms three points to three points via affine transformation
        """
        # points to be transformed
        p = np.array([[x1, y1, 1.0],
                   [x2, y2, 1.0],
                   [x3, y3, 1.0]])
        # destination
        d = np.array([[x1d, y1d, 1.0],
                   [x2d, y2d, 1.0],
                   [x3d, y3d, 1.0]])
        # transformation
        t = np.dot(linalg.inv(p), d)
        return t


if __name__ == '__main__':
    """
    testing
    """
    print("#### TEST 1 ######")
    FourPoint(1, 3, 2, 2, 2, 4, 4, 3, 1, 3, 2, 2, 4, 3, 2, 4)
    print("#### TEST 2 ######")
    FourPoint(1, 3, 1, 1, 2, 2, 2, 1, 1, 3, 2, 2, 4, 3, 2, 4)
    print("#### TEST 3######")
    FourPoint(1, 4, 1, 2, 2, 2, 2, 1, 1, 3, 2, 2, 4, 3, 2, 4)
    print("#### TEST 4######")
    FourPoint(1, 4, 1, 2, 2, 1, 2, 2, 1, 3, 2, 2, 4, 3, 2, 4)
    print(FourPoint(1, 4, 1, 2, 2, 1, 2, 2, 1, 3, 2, 2, 4, 3, 2, 4).calc_distance_points(1, 1, 2, 2))
    #print(FourPoint.calc_distance_points(1, 1, 2, 2))
